from __future__ import print_function
import numpy as np

np.random.seed(1337)
# 生产环境
from aa_deep_learning.aadeeplearning import AADeepLearning
from aa_deep_learning.aadeeplearning.datasets import mnist
from aa_deep_learning.aadeeplearning.datasets import np_utils

# 测试环境
# from aadeeplearning import aadeeplearning as aa

# 输入图像的维度，此处是mnist图像，因此是28*28
img_rows, img_cols = 28, 28
# 10分类
nb_classes = 10

# keras中的mnist数据集已经被划分成了60,000个训练集，10,000个测试集的形式，按以下格式调用即可
(X_train, y_train), (X_test, y_test) = mnist.load_data()
X_test = X_test[:64]
y_test = y_test[:64]

# print(X_test[0])

# 画出minist 数字
# import matplotlib.pyplot as plt
# fig = plt.figure()
# plt.imshow(X_test[0],cmap = 'binary')#黑白显示
# plt.show()

# 后端使用tensorflow时，即tf模式下，
# 会将100张RGB三通道的16*32彩色图表示为(100,16,32,3)，
# 第一个维度是样本维，表示样本的数目，
# 第二和第三个维度是高和宽，
# 最后一个维度是通道维，表示颜色通道数
X_train = X_train.reshape(X_train.shape[0], img_rows, img_cols, 1)
X_test = X_test.reshape(X_test.shape[0], img_rows, img_cols, 1)
input_shape = (img_rows, img_cols, 1)

# 将X_train, X_test的数据格式转为float32
X_train = X_train.astype('float32')
X_test = X_test.astype('float32')
# 归一化
X_train /= 255
X_test /= 255
# 打印出相关信息
print('X_train shape:', X_train.shape)
print('y_train shape:', y_train.shape)
print('X_test shape:', X_test.shape)
print('y_test shape:', y_test.shape)
print(X_train.shape[0], 'train samples')
print(X_test.shape[0], 'test samples')
# X_train shape: (60000, 28, 28, 1)
# 60000 train samples
# 10000 test samples

# 将类别向量(从0到nb_classes的整数向量)映射为二值类别矩阵，
# 相当于将向量用one-hot重新编码
Y_train = np_utils.to_categorical(y_train, nb_classes)
Y_test = np_utils.to_categorical(y_test, nb_classes)
# print(Y_test.shape)
# 网络配置文件
solver = {
    # 初始学习率
    "learning_rate": 0.001,
    # 学习率衰减: 通常设置为 0.99
    "learning_rate_decay": 0.9999,
    # 优化策略: sgd/momentum/rmsprop
    "optimizer": "momentum",
    # 如果想使用添加动量的梯度下降算法做优化,需要设置这一项，通常设置为 0.9/0.95 即可，一般不需要调整
    "momentum_coefficient": 0.95,
    # rmsprop优化器的衰减系数
    "rmsprop_decay": 0.95,
    # 正则化系数
    "reg_coefficient": 0,
    # 训练多少次
    "number_iteration": 8000,
    # 每次用多少个样本训练
    "batch_size": 64,
    # 每隔几个迭代周期评估一次准确率？
    "evaluate_interval": 10,
    # 是否更新学习率？  true/false
    "learning_rate_update": True,
    # 是否需要保存模型快照？ true/false
    "snapshot": False,
    # 每隔几个迭代周期保存一次快照？
    "snapshot_interval": 5,
    # 是否以fine_tune方式训练？ true/false
    "fine_tune": False,
    # 预训练参数模型所在路径
    "pre_train_model": "./iter5.gordonmodel"
}

net = [
    {
        # 层名
        "name": "convolutional_1",
        # 层类型
        "type": "convolutional",
        # 卷积核个数
        "kernel_number": 32,
        # 卷积核高
        "kernel_height": 5,
        # 卷积核宽
        "kernel_width": 5,
        # 填充数
        "padding": 1,
        # 滑动步长
        "stride": 1,
        # 权重初始化  gaussian/msra
        "weight_init": "msra"
    },
    {
        # 层名
        "name": "pooling_1",
        # 层类型（最大池化）
        "type": "pooling",
        # 池化核高
        "kernel_height": 3,
        # 池化核宽
        "kernel_width": 3,
        # 滑动步长
        "stride": 2
    },
    {
        # 层名
        "name": "relu_1",
        # 层类型
        "type": "relu"
    },
    {
        # 层名
        "name": "convolutional_2",
        # 层类型
        "type": "convolutional",
        # 卷积核个数
        "kernel_number": 32,
        # 卷积核高
        "kernel_height": 5,
        # 卷积核宽
        "kernel_width": 5,
        # 填充数
        "padding": 1,
        # 滑动步长
        "stride": 1,
        # 权重初始化  gaussian/msra
        "weight_init": "msra"
    },
    {
        # 层名
        "name": "relu_2",
        # 层类型
        "type": "relu"
    },
    {
        # 层名
        "name": "pooling_2",
        # 层类型
        "type": "pooling",
        # 池化核高
        "kernel_height": 3,
        # 池化核宽
        "kernel_width": 3,
        # 滑动步长
        "stride": 2
    },
    {
        # 层名
        "name": "fully_connected_1",
        # 层类型
        "type": "fully_connected",
        # 神经元个数
        "neurons_number": 256,
        # 权重初始化方式  msra/gaussian
        "weight_init": "msra"
    },
    {
        # 层名
        "name": "relu_1",
        # 层类型 可选，relu，sigmoid，tanh，
        "type": "relu",
        # relu函数的上界限制值；   注： 不做设置或者设置为0意味着是普通的relu函数
        "threshold": 0
    },
    # {
    #     # 层名
    #     "name": "drop1",
    #     # 层类型
    #     "type": "Dropout",
    #     # dropout比例（置零比例）
    #     "drop rate": 0.5
    # },
    {
        # 层名
        "name": "fully_connected_2",
        # 层类型
        "type": "fully_connected",
        # 神经元个数
        "neurons_number": 128,
        # 权重初始化方式  msra/gaussian
        "weight_init": "msra"
    },
    {
        # 层名
        "name": "relu_2",
        # 层类型
        "type": "relu",
        # relu函数的上界限制值；   注： 不做设置或者设置为0意味着是普通的relu函数
        "threshold": 0
    },
    {
        # 层名
        "name": "fully_connected_3",
        # 层类型
        "type": "fully_connected",
        # 神经元个数
        "neurons_number": 10,
        # 权重初始化方式  msra/gaussian
        "weight_init": "msra"
    },
    {
        # 层名
        "name": "softmax",
        # 层类型
        "type": "softmax"
    }
]

AA = AADeepLearning(solver, net)
# 训练模型
AA.train(X_train, Y_train, X_test, Y_test)

# 按batch计算在某些输入数据上模型的误差
score = AA.predict(X_test, Y_test)

"""
# 输出训练好的模型在测试集上的表现
print('Test score:', score[0])
print('Test accuracy:', score[1])

# Test score: 0.032927570413
# Test accuracy: 0.9892
"""



from __future__ import print_function
import numpy as np
np.random.seed(1337)

from keras.datasets import mnist
from keras.models import Sequential
from keras.layers import Dense, Dropout, Activation, Flatten
from keras.layers import Conv2D, MaxPooling2D
from keras.utils import np_utils

batch_size = 128
nb_classes = 10
nb_epoch = 12


# 输入图像的维度，此处是mnist图像，因此是28*28
img_rows, img_cols = 28, 28
# 卷积层中使用的卷积核的个数
nb_filters = 32
# 池化层操作的范围
pool_size = (2,2)
# 卷积核的大小
kernel_size = (3,3)
# keras中的mnist数据集已经被划分成了60,000个训练集，10,000个测试集的形式，按以下格式调用即可
(X_train, y_train), (X_test, y_test) = mnist.load_data()

# 后端使用tensorflow时，即tf模式下，
# 会将100张RGB三通道的16*32彩色图表示为(100,16,32,3)，
# 第一个维度是样本维，表示样本的数目，
# 第二和第三个维度是高和宽，
# 最后一个维度是通道维，表示颜色通道数
X_train = X_train.reshape(X_train.shape[0], img_rows, img_cols, 1)
X_test = X_test.reshape(X_test.shape[0], img_rows, img_cols, 1)
input_shape = (img_rows, img_cols, 1)

# 将X_train, X_test的数据格式转为float32
X_train = X_train.astype('float32')
X_test = X_test.astype('float32')
# 归一化
X_train /= 255
X_test /= 255
# 打印出相关信息
print('X_train shape:', X_train.shape)
print(X_train.shape[0], 'train samples')
print(X_test.shape[0], 'test samples')

# X_train shape: (60000, 28, 28, 1)
# 60000 train samples
# 10000 test samples

# 将类别向量(从0到nb_classes的整数向量)映射为二值类别矩阵，
# 相当于将向量用one-hot重新编码
Y_train = np_utils.to_categorical(y_train, nb_classes)
Y_test = np_utils.to_categorical(y_test, nb_classes)

# 建立序贯模型
model = Sequential()

# 卷积层，对二维输入进行滑动窗卷积
# 当使用该层为第一层时，应提供input_shape参数，在tf模式中，通道维位于第三个位置
# border_mode：边界模式，为"valid","same"或"full"，即图像外的边缘点是补0
# 还是补成相同像素，或者是补1
layer = Conv2D(nb_filters, kernel_size,activation='relu',input_shape=input_shape)
# print(type(layer))
# exit()
model.add(layer)


# 卷积层，激活函数是ReLu
# model.add(Conv2D(nb_filters, kernel_size))
model.add(Activation('relu'))

# 池化层，选用Maxpooling，给定pool_size，dropout比例为0.25
model.add(MaxPooling2D(pool_size=pool_size))
model.add(Dropout(0.25))

# Flatten层，把多维输入进行一维化，常用在卷积层到全连接层的过渡
model.add(Flatten())

# 包含128个神经元的全连接层，激活函数为ReLu，dropout比例为0.5
model.add(Dense(128))
model.add(Activation('relu'))
model.add(Dropout(0.5))

# 包含10个神经元的输出层，激活函数为Softmax
model.add(Dense(nb_classes))
model.add(Activation('softmax'))

# 输出模型的参数信息
# model.summary()
# 配置模型的学习过程
model.compile(loss='categorical_crossentropy',
              optimizer='adadelta',
              metrics=['accuracy'])

# 训练模型
model.fit(X_train, Y_train, batch_size=batch_size, epochs=nb_epoch,
          verbose=1, validation_data=(X_test, Y_test))

# 按batch计算在某些输入数据上模型的误差
score = model.evaluate(X_test, Y_test, verbose=0)

# 输出训练好的模型在测试集上的表现
print('Test score:', score[0])
print('Test accuracy:', score[1])

# Test score: 0.032927570413
# Test accuracy: 0.9892


"""Contains the base Layer class, from which all layers inherit.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy as np


class Relu():
    @staticmethod
    def init(layer):
        return layer
    @staticmethod
    def forword(flow_data):
        return np.maximum(0, flow_data)
    @staticmethod
    def backword(flow_data, layer):
        # drelu/dz  = 小于零就始终等于0 ，大于0就等于一
        return flow_data * np.array(layer["output"] > 0)



"""Contains the base Layer class, from which all layers inherit.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

class Layer(object):
    def __init__(self):
        pass
    def init(self, neural_net,input_shape):
        pass
    def forward(self):
        pass
    def backward(self):
        pass



# -*- coding: UTF-8 -*-
'''
Created by huxiaoman 2017.11.22
'''
import numpy as np

class ReluActivator(object):
    def forward(self, weighted_input):
        #return weighted_input
        return max(0, weighted_input)

    def backward(self, output):
        return 1 if output > 0 else 0


class IdentityActivator(object):
    def forward(self, weighted_input):
        return weighted_input

    def backward(self, output):
        return 1


class SigmoidActivator(object):
    def forward(self, weighted_input):
        return 1.0 / (1.0 + np.exp(-weighted_input))
    #the partial of sigmoid
    def backward(self, output):
        return output * (1 - output)


class TanhActivator(object):
    def forward(self, weighted_input):
        return 2.0 / (1.0 + np.exp(-2 * weighted_input)) - 1.0

    def backward(self, output):
        return 1 - output * output


# 获取卷积区域
def get_patch(input_array, i, j, filter_width,
              filter_height, stride):
    '''
    从输入数组中获取本次卷积的区域，
    自动适配输入为2D和3D的情况
    '''
    start_i = i * stride
    start_j = j * stride
    if input_array.ndim == 2:
        input_array_conv = input_array[
                           start_i: start_i + filter_height,
                           start_j: start_j + filter_width]
        return input_array_conv
    elif input_array.ndim == 3:
        input_array_conv = input_array[:,
                           start_i: start_i + filter_height,
                           start_j: start_j + filter_width]
        return input_array_conv


# 获取一个2D区域的最大值所在的索引
def get_max_index(array):
    max_i = 0
    max_j = 0
    max_value = array[0, 0]
    for i in range(array.shape[0]):
        for j in range(array.shape[1]):
            if array[i, j] > max_value:
                max_value = array[i, j]
                max_i, max_j = i, j
    return max_i, max_j


# 计算卷积
def conv(input_array,
         kernel_array,
         output_array,
         stride, bias):
    '''
    计算卷积，自动适配输入为2D和3D的情况
    '''
    channel_number = input_array.ndim
    output_width = output_array.shape[1]
    output_height = output_array.shape[0]
    kernel_width = kernel_array.shape[-1]
    kernel_height = kernel_array.shape[-2]
    for i in range(output_height):
        for j in range(output_width):
            output_array[i][j] = (
                                         get_patch(input_array, i, j, kernel_width,
                                                   kernel_height, stride) * kernel_array
                                 ).sum() + bias


# 为数组增加Zero padding
def padding(input_array, zp):
    '''
    为数组增加Zero padding，自动适配输入为2D和3D的情况
    '''
    if zp == 0:
        return input_array
    else:
        if input_array.ndim == 3:
            input_width = input_array.shape[2]
            input_height = input_array.shape[1]
            input_depth = input_array.shape[0]
            padded_array = np.zeros((
                input_depth,
                input_height + 2 * zp,
                input_width + 2 * zp))
            padded_array[:,
            zp: zp + input_height,
            zp: zp + input_width] = input_array
            return padded_array
        elif input_array.ndim == 2:
            input_width = input_array.shape[1]
            input_height = input_array.shape[0]
            padded_array = np.zeros((
                input_height + 2 * zp,
                input_width + 2 * zp))
            padded_array[zp: zp + input_height,
            zp: zp + input_width] = input_array
            return padded_array


# 对numpy数组进行element wise操作
def element_wise_op(array, op):
    for i in np.nditer(array,
                       op_flags=['readwrite']):
        i[...] = op(i)


class Filter(object):
    def __init__(self, width, height, depth):
        self.weights = np.random.uniform(-1e-4, 1e-4,
                                         (depth, height, width))
        self.bias = 0
        self.weights_grad = np.zeros(
            self.weights.shape)
        self.bias_grad = 0

    def __repr__(self):
        return 'filter weights:\n%s\nbias:\n%s' % (
            repr(self.weights), repr(self.bias))

    def get_weights(self):
        return self.weights

    def get_bias(self):
        return self.bias

    def update(self, learning_rate):
        self.weights -= learning_rate * self.weights_grad
        self.bias -= learning_rate * self.bias_grad


class ConvLayer(object):
    def __init__(self, input_width, input_height,
                 channel_number, filter_width,
                 filter_height, filter_number,
                 zero_padding, stride, activator,
                 learning_rate):
        self.input_width = input_width
        self.input_height = input_height
        self.channel_number = channel_number
        self.filter_width = filter_width
        self.filter_height = filter_height
        self.filter_number = filter_number
        self.zero_padding = zero_padding
        self.stride = stride
        self.output_width = \
            ConvLayer.calculate_output_size(
                self.input_width, filter_width, zero_padding,
                stride)
        self.output_height = \
            ConvLayer.calculate_output_size(
                self.input_height, filter_height, zero_padding,
                stride)
        self.output_array = np.zeros((self.filter_number,
                                      self.output_height, self.output_width))
        self.filters = []
        for i in range(filter_number):
            self.filters.append(Filter(filter_width,
                                       filter_height, self.channel_number))
        self.activator = activator
        self.learning_rate = learning_rate

    def forward(self, input_array):
        '''
        计算卷积层的输出
        输出结果保存在self.output_array
        '''
        self.input_array = input_array
        self.padded_input_array = padding(input_array,
                                          self.zero_padding)
        for f in range(self.filter_number):
            filter = self.filters[f]
            conv(self.padded_input_array,
                 filter.get_weights(), self.output_array[f],
                 self.stride, filter.get_bias())
        element_wise_op(self.output_array,
                        self.activator.forward)

    def backward(self, input_array, sensitivity_array,
                 activator):
        '''
        计算传递给前一层的误差项，以及计算每个权重的梯度
        前一层的误差项保存在self.delta_array
        梯度保存在Filter对象的weights_grad
        '''
        self.forward(input_array)
        self.bp_sensitivity_map(sensitivity_array,
                                activator)
        self.bp_gradient(sensitivity_array)

    def update(self):
        '''
        按照梯度下降，更新权重
        '''
        for filter in self.filters:
            filter.update(self.learning_rate)

    def bp_sensitivity_map(self, sensitivity_array,
                           activator):
        '''
        计算传递到上一层的sensitivity map
        sensitivity_array: 本层的sensitivity map
        activator: 上一层的激活函数
        '''
        # 处理卷积步长，对原始sensitivity map进行扩展
        expanded_array = self.expand_sensitivity_map(
            sensitivity_array)
        # full卷积，对sensitivitiy map进行zero padding
        # 虽然原始输入的zero padding单元也会获得残差
        # 但这个残差不需要继续向上传递，因此就不计算了
        expanded_width = expanded_array.shape[2]
        zp = (self.input_width +
              self.filter_width - 1 - expanded_width) / 2
        padded_array = padding(expanded_array, zp)
        # 初始化delta_array，用于保存传递到上一层的
        # sensitivity map
        self.delta_array = self.create_delta_array()
        # 对于具有多个filter的卷积层来说，最终传递到上一层的
        # sensitivity map相当于所有的filter的
        # sensitivity map之和
        for f in range(self.filter_number):
            filter = self.filters[f]
            # 将filter权重翻转180度
            flipped_weights = np.array(map(
                lambda i: np.rot90(i, 2),
                filter.get_weights()))
            # 计算与一个filter对应的delta_array
            delta_array = self.create_delta_array()
            for d in range(delta_array.shape[0]):
                conv(padded_array[f], flipped_weights[d],
                     delta_array[d], 1, 0)
            self.delta_array += delta_array
        # 将计算结果与激活函数的偏导数做element-wise乘法操作
        derivative_array = np.array(self.input_array)
        element_wise_op(derivative_array,
                        activator.backward)
        self.delta_array *= derivative_array

    def bp_gradient(self, sensitivity_array):
        # 处理卷积步长，对原始sensitivity map进行扩展
        expanded_array = self.expand_sensitivity_map(
            sensitivity_array)
        for f in range(self.filter_number):
            # 计算每个权重的梯度
            filter = self.filters[f]
            for d in range(filter.weights.shape[0]):
                conv(self.padded_input_array[d],
                     expanded_array[f],
                     filter.weights_grad[d], 1, 0)
            # 计算偏置项的梯度
            filter.bias_grad = expanded_array[f].sum()

    def expand_sensitivity_map(self, sensitivity_array):
        depth = sensitivity_array.shape[0]
        # 确定扩展后sensitivity map的大小
        # 计算stride为1时sensitivity map的大小
        expanded_width = (self.input_width -
                          self.filter_width + 2 * self.zero_padding + 1)
        expanded_height = (self.input_height -
                           self.filter_height + 2 * self.zero_padding + 1)
        # 构建新的sensitivity_map
        expand_array = np.zeros((depth, expanded_height,
                                 expanded_width))
        # 从原始sensitivity map拷贝误差值
        for i in range(self.output_height):
            for j in range(self.output_width):
                i_pos = i * self.stride
                j_pos = j * self.stride
                expand_array[:, i_pos, j_pos] = \
                    sensitivity_array[:, i, j]
        return expand_array

    def create_delta_array(self):
        return np.zeros((self.channel_number,
                         self.input_height, self.input_width))

    @staticmethod
    def calculate_output_size(input_size,
                              filter_size, zero_padding, stride):
        return (input_size - filter_size +
                2 * zero_padding) / stride + 1


class MaxPoolingLayer(object):
    def __init__(self, input_width, input_height,
                 channel_number, filter_width,
                 filter_height, stride):
        self.input_width = input_width
        self.input_height = input_height
        self.channel_number = channel_number
        self.filter_width = filter_width
        self.filter_height = filter_height
        self.stride = stride
        self.output_width = (input_width -
                             filter_width) / self.stride + 1
        self.output_height = (input_height -
                              filter_height) / self.stride + 1
        self.output_array = np.zeros((self.channel_number,
                                      self.output_height, self.output_width))

    def forward(self, input_array):
        for d in range(self.channel_number):
            for i in range(self.output_height):
                for j in range(self.output_width):
                    self.output_array[d, i, j] = (
                        get_patch(input_array[d], i, j,
                                  self.filter_width,
                                  self.filter_height,
                                  self.stride).max())

    def backward(self, input_array, sensitivity_array):
        self.delta_array = np.zeros(input_array.shape)
        for d in range(self.channel_number):
            for i in range(self.output_height):
                for j in range(self.output_width):
                    patch_array = get_patch(
                        input_array[d], i, j,
                        self.filter_width,
                        self.filter_height,
                        self.stride)
                    k, l = get_max_index(patch_array)
                    self.delta_array[d,
                                     i * self.stride + k,
                                     j * self.stride + l] = \
                        sensitivity_array[d, i, j]


###测试前向传播和反向传播后跟新的filter结果是否正确
def init_test():
    a = np.array(
        [[[0, 1, 1, 0, 2],
          [2, 2, 2, 2, 1],
          [1, 0, 0, 2, 0],
          [0, 1, 1, 0, 0],
          [1, 2, 0, 0, 2]],
         [[1, 0, 2, 2, 0],
          [0, 0, 0, 2, 0],
          [1, 2, 1, 2, 1],
          [1, 0, 0, 0, 0],
          [1, 2, 1, 1, 1]],
         [[2, 1, 2, 0, 0],
          [1, 0, 0, 1, 0],
          [0, 2, 1, 0, 1],
          [0, 1, 2, 2, 2],
          [2, 1, 0, 0, 1]]])
    b = np.array(
        [[[0, 1, 1],
          [2, 2, 2],
          [1, 0, 0]],
         [[1, 0, 2],
          [0, 0, 0],
          [1, 2, 1]]])
    cl = ConvLayer(5, 5, 3, 3, 3, 2, 1, 2, IdentityActivator(), 0.001)
    cl.filters[0].weights = np.array(
        [[[-1, 1, 0],
          [0, 1, 0],
          [0, 1, 1]],
         [[-1, -1, 0],
          [0, 0, 0],
          [0, -1, 0]],
         [[0, 0, -1],
          [0, 1, 0],
          [1, -1, -1]]], dtype=np.float64)
    cl.filters[0].bias = 1
    cl.filters[1].weights = np.array(
        [[[1, 1, -1],
          [-1, -1, 1],
          [0, -1, 1]],
         [[0, 1, 0],
          [-1, 0, -1],
          [-1, 1, 0]],
         [[-1, 0, 0],
          [-1, 0, 1],
          [-1, 0, 0]]], dtype=np.float64)
    return a, b, cl


def test():
    a, b, cl = init_test()
    cl.forward(a)
    print
    "前向传播结果:", cl.output_array
    cl.backward(a, b, IdentityActivator())
    cl.update()
    print
    "反向传播后更新得到的filter1:", cl.filters[0]
    print
    "反向传播后更新得到的filter2:", cl.filters[1]


def gradient_check():
    '''
    梯度检查
    '''
    # 设计一个误差函数，取所有节点输出项之和
    error_function = lambda o: o.sum()

    # 计算forward值
    a, b, cl = init_test()
    cl.forward(a)

    # 求取sensitivity map
    sensitivity_array = np.ones(cl.output_array.shape,
                                dtype=np.float64)
    # 计算梯度
    cl.backward(a, sensitivity_array,
                IdentityActivator())
    # 检查梯度
    epsilon = 10e-4
    for d in range(cl.filters[0].weights_grad.shape[0]):
        for i in range(cl.filters[0].weights_grad.shape[1]):
            for j in range(cl.filters[0].weights_grad.shape[2]):
                cl.filters[0].weights[d, i, j] += epsilon
                cl.forward(a)
                err1 = error_function(cl.output_array)
                cl.filters[0].weights[d, i, j] -= 2 * epsilon
                cl.forward(a)
                err2 = error_function(cl.output_array)
                expect_grad = (err1 - err2) / (2 * epsilon)
                cl.filters[0].weights[d, i, j] += epsilon
                print
                'weights(%d,%d,%d): expected - actural %f - %f' % (
                    d, i, j, expect_grad, cl.filters[0].weights_grad[d, i, j])


def init_pool_test():
    a = np.array(
        [[[1, 1, 2, 4],
          [5, 6, 7, 8],
          [3, 2, 1, 0],
          [1, 2, 3, 4]],
         [[0, 1, 2, 3],
          [4, 5, 6, 7],
          [8, 9, 0, 1],
          [3, 4, 5, 6]]], dtype=np.float64)

    b = np.array(
        [[[1, 2],
          [2, 4]],
         [[3, 5],
          [8, 2]]], dtype=np.float64)

    mpl = MaxPoolingLayer(4, 4, 2, 2, 2, 2)

    return a, b, mpl


def test_pool():
    a, b, mpl = init_pool_test()
    mpl.forward(a)
    print
    'input array:\n%s\noutput array:\n%s' % (a, mpl.output_array)
    mpl.backward(a, b)
    print
    'input array:\n%s\nsensitivity array:\n%s\ndelta array:\n%s' % (a, b, mpl.delta_array)


def test_example():
    a = np.array(
        [[1, 0, 1, 0],
         [1, 1, 1, 0],
         [1, 0, 1, 0],
         [0, 0, 1, 0]],
        dtype=np.float64)
    b = np.array(
        [[[1, -1],
          [1, -1]],
         [[1, 1],
          [-1, -1]]], dtype=np.float64)


if __name__ == "__main__":
    test()


#fc_net = [12288,40,40,40,40,1] 
#交叉熵：迭代6001次，学习率0.01（适合），除了第一层，其它层随着层级递增，a值向0聚拢，同样使得梯度消失
#交叉熵：加入BN后，迭代6001次，学习率0.01（适合），随着层级递增，分布都很均匀，有利于训练！  

#交叉熵：迭代6001次，学习率0.5（过大），随着迭代次数增多，第一层激活值会趋于1和-1（激活函数梯度趋于0） ---->z值向0的两侧扩散，梯度消失
#交叉熵：加入BN后，迭代6001次，学习率0.5（过大），随着迭代次数增多，第一层激活值会缓慢趋于1和-1 ---->z值向0的两侧扩散，梯度消失有所减缓
import time
    start = time.clock()
    parameters = Train_Net(fc_net,train_set_x,train_set_y,iterations=200,LearnRate=0.5,bn_mode="train",use_bn=False,active_func="tanh")
    end = time.clock()
    print ("time cost=",end-start,"s")







import h5py
import matplotlib.pyplot as plt
import numpy as np

#字典！以键值对的方式保存的一种数据结构
def load_dataset():
    #1.创建文件对象
    train_dataset = h5py.File('./我的数据集/train_catvnoncat.h5','r')
    test_dataset = h5py.File('./我的数据集/test_catvnoncat.h5','r')
    #2.读取数据
    train_set_x = np.array(train_dataset["train_set_x"][:])
    train_set_y = np.array(train_dataset["train_set_y"][:])
    test_set_x = np.array(test_dataset["test_set_x"][:])
    test_set_y = np.array(test_dataset["test_set_y"][:])
    plt.figure(figsize=(2,2))
    plt.imshow(train_set_x[110])
    plt.show()
    #3.变化维度以适应神经网络输入
    #(209, 64, 64, 3) -> (12288,209)
    train_set_x = train_set_x.reshape(train_set_x.shape[0],-1).T  #(12288,209)
    test_set_x = test_set_x.reshape(test_set_x.shape[0],-1).T  #(12288,209)
    #(209,)->(1,209)
    train_set_y = train_set_y.reshape(1,train_set_y.shape[0]) #(1,209)
    test_set_y = test_set_y.reshape(1,test_set_y.shape[0]) #(1,209)
    
    return train_set_x,train_set_y,test_set_x,test_set_y

def init_parameters(fc_net):
    parameters = {}
    Layer_num = len(fc_net) #Layer_num=5    
    for L in range(1,Layer_num):
        parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*0.01   #最普通的随机初始化
        #parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])* np.sqrt(1/fc_net[L-1])
        parameters["b"+str(L)] = np.zeros((fc_net[L],1))  #使用BN的时候其实可以去掉偏置b了（要在很多代码里面改，先不改了......）
        parameters["gamma"+str(L)] = np.ones((fc_net[L],1))   
        parameters["beta"+str(L)] = np.zeros((fc_net[L],1))   
        parameters["running_mean"+str(L)] = np.zeros((fc_net[L],1))   
        parameters["running_std"+str(L)] = np.zeros((fc_net[L],1))         
    for L in range(1,Layer_num):
        print("W"+str(L)+"=",parameters["W"+str(L)].shape)
        print("b"+str(L)+"=",parameters["b"+str(L)].shape)                       
    return parameters

def init_parameters_momentum(fc_net):
    parameters = {}
    Layer_num = len(fc_net) #Layer_num=5 
    v = {}
    for L in range(1,Layer_num):
        #parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*0.01  #
        parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*np.sqrt(1/fc_net[L-1]) #Xavier初始化，针对tanh函数
        #parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*np.sqrt(2/fc_net[L-1]) #He初始化，针对ReLU函数
        parameters["b"+str(L)] = np.zeros((fc_net[L],1))     
        v["dW"+str(L)] = np.zeros((fc_net[L],fc_net[L-1]))
        v["db"+str(L)] = np.zeros((fc_net[L],1))           
    for L in range(1,Layer_num):
        print("W"+str(L)+"=",parameters["W"+str(L)].shape)
        print("b"+str(L)+"=",parameters["b"+str(L)].shape)                       
    return parameters,v

def init_parameters_RMSProp(fc_net):
    parameters = {}
    Layer_num = len(fc_net) #Layer_num=5 
    s = {}
    for L in range(1,Layer_num):
        #parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*0.01  #
        parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*np.sqrt(1/fc_net[L-1]) #Xavier初始化，针对tanh函数
        #parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*np.sqrt(2/fc_net[L-1]) #He初始化，针对ReLU函数
        parameters["b"+str(L)] = np.zeros((fc_net[L],1))     
        s["dW"+str(L)] = np.zeros((fc_net[L],fc_net[L-1]))
        s["db"+str(L)] = np.zeros((fc_net[L],1))           
    for L in range(1,Layer_num):
        print("W"+str(L)+"=",parameters["W"+str(L)].shape)
        print("b"+str(L)+"=",parameters["b"+str(L)].shape)                       
    return parameters,s

def ReLU(Z):
    return np.maximum(0,Z)

def tanh(Z):
    return np.tanh(Z)

def sigmoid(Z):
    return 1/(1+np.exp(-Z))

def BN_forward(Z,bn_mode,parameters,L,epsilon=1e-9,mu=0.9):
    gamma,beta = parameters["gamma"+str(L)],parameters["beta"+str(L)]
	Z_mean,Z_std = np.zeros((Z.shape[0],1)),np.zeros((Z.shape[0],1))  
    Z_normed,Z_out = np.zeros(Z.shape),np.zeros(Z.shape)
     
    if bn_mode == 'train':       
        Z_mean = np.mean(Z, axis=1, keepdims=True)#计算均值  
        Z_std = np.std(Z, axis=1, keepdims=True)  #计算标准差  
        Z_normed = (Z - Z_mean) / (Z_std + epsilon)  #归一化  
        Z_out = gamma * Z_normed + beta    #重构变换
        parameters["running_mean"+str(L)] = mu * parameters["running_mean"+str(L)] + (1 - mu) * Z_mean
        parameters["running_std"+str(L)] = mu * parameters["running_std"+str(L)] + (1 - mu) * Z_std
    elif bn_mode == 'test':        
        Z_normed = (Z - parameters["running_mean"+str(L)]) / (parameters["running_std"+str(L)] + epsilon)
        Z_out = gamma * Z_normed + beta
    else:
        raise ValueError('Invalid forward batchnorm mode "%s"' % mode)
    return Z_out,Z_mean,Z_std,Z_normed,parameters

def BN_backward(dZ,bn_cache,parameters,L,epsilon=1e-9):
    m = dZ.shape[1] #样本数
    Z_mean,Z_std,Z_normed = bn_cache["Z_mean"+str(L)],bn_cache["Z_std"+str(L)],bn_cache["Z_normed"+str(L)]
    gamma,beta = parameters["gamma"+str(L)],parameters["beta"+str(L)]  
    dgamma = np.sum(dZ * Z_normed,axis=1, keepdims=True)            
    dbeta = np.sum(dZ,axis=1, keepdims=True)
    dZ_out = (gamma/(Z_std+epsilon)) * (dZ - dgamma*Z_normed/m - np.mean(dZ,axis=1,keepdims=True))
    return dZ_out, dgamma, dbeta

def forward_pass(A0,parameters,bn_mode,use_bn=True,active_func="ReLU"):    #前向计算函数
    cache = {}
    bn_cache = {}
    A = A0
    cache["A0"] = A0
    Layer_num = len(parameters) // 6 #写成/2代表精确除法，得到一个浮点数，写成//2代表向下取整除法，得到一个整数
   
    for L in range(1,Layer_num): #遍历[1,2,3]
        #1.求和
        if use_bn:
            Z = np.dot(parameters["W"+str(L)],A)
        else:
            Z = np.dot(parameters["W"+str(L)],A) + parameters["b"+str(L)]
        
        #2.批归一化
        if(use_bn):
            Z, Z_mean, Z_std, Z_normed,parameters = BN_forward(Z,bn_mode,parameters,L)
            bn_cache["Z_mean"+str(L)] = Z_mean
            bn_cache["Z_std"+str(L)] = Z_std
            bn_cache["Z_normed"+str(L)] = Z_normed
        #3.激活
        if active_func=="sigmoid":
            A = sigmoid(Z)
        elif active_func=="tanh":
            A = tanh(Z)
        else:
            A = ReLU(Z)
        cache["A"+str(L)] = A
        cache["Z"+str(L)] = Z
        
    Z = np.dot(parameters["W"+str(Layer_num)],A) + parameters["b"+str(Layer_num)] 
    A = sigmoid(Z)
    cache["A"+str(Layer_num)] = A
    cache["Z"+str(Layer_num)] = Z
    return A,cache,bn_cache,parameters


def compute_cost(AL,Y):
    m = Y.shape[1]   #Y =（1,209）
    cost = - np.sum(  Y * np.log(AL) + (1 - Y) * np.log(1 - AL)  ) / m  #交叉熵
    #cost = (1/m)*np.sum((1/2)*(AL-Y)*(AL-Y)) #代价函数
    return cost

def backward_pass(AL,parameters,cache,bn_cache,Y,use_bn=False,active_func="ReLU"):
    m = Y.shape[1]  #样本总数
    gradient = {}   #保存各层参数梯度的字典
    Layer_num = len(parameters) // 6
    dZL = - (np.divide(Y, AL) - np.divide(1 - Y, 1 - AL)) * AL*(1-AL)  #交叉熵代价函数dZ
    #dZL = (AL-Y)*(AL*(1-AL))#获取最末层误差信号 dZL.shape = (1,209)  #最后一层无需做BN处理
    gradient["dW"+str(Layer_num)] = (1/m)*np.dot(dZL,cache["A"+str(Layer_num-1)].T)
    gradient["db"+str(Layer_num)] = (1/m)*np.sum(dZL,axis=1,keepdims = True)    
    for L in reversed(range(1,Layer_num)): #遍历[3,2,1]，其中reversed函数[1,2,3]颠倒为[3,2,1]
        if active_func=="sigmoid":
            dZL = np.dot(parameters["W"+str(L+1)].T,dZL)*(cache["A"+str(L)]*(1-cache["A"+str(L)]))  
        elif active_func=="tanh":
            dZL = np.dot(parameters["W"+str(L+1)].T,dZL)*(1-np.power(cache["A"+str(L)],2))#dtanh/dz  = 1-a^2
        else:
            dZL = np.dot(parameters["W"+str(L+1)].T,dZL)* np.array(cache["Z"+str(L)]>0)           
        if(use_bn):
            dZL, dgamma, dbeta = BN_backward(dZL,bn_cache,parameters,L)
            gradient["dgamma"+str(L)] = dgamma
            gradient["dbeta"+str(L)] = dbeta       
        gradient["dW"+str(L)] = (1/m)*np.dot(dZL,cache["A"+str(L-1)].T)
        if use_bn==False:
            gradient["db"+str(L)] = (1/m)*np.sum(dZL,axis=1,keepdims = True)    
    return gradient

def update_parameters(gradient,parameters,LearnRate,use_bn=False):
    Layer_num = len(parameters) // 6
    if(use_bn):
        for L in range(1,Layer_num): #遍历[1,2,3] ,因为最后一层不做BatchNormalization
            parameters["W"+str(L)] = parameters["W"+str(L)] - LearnRate*gradient["dW"+str(L)]
            parameters["gamma"+str(L)] = parameters["gamma"+str(L)] - LearnRate*gradient["dgamma"+str(L)]
            parameters["beta"+str(L)] = parameters["beta"+str(L)] - LearnRate*gradient["dbeta"+str(L)]  
        parameters["W"+str(Layer_num)] = parameters["W"+str(Layer_num)] - LearnRate*gradient["dW"+str(Layer_num)]
        parameters["b"+str(Layer_num)] = parameters["b"+str(Layer_num)] - LearnRate*gradient["db"+str(Layer_num)]  
    else:
        for L in range(1,Layer_num+1): #遍历[1,2,3,4]
            parameters["W"+str(L)] = parameters["W"+str(L)] - LearnRate*gradient["dW"+str(L)]
            parameters["b"+str(L)] = parameters["b"+str(L)] - LearnRate*gradient["db"+str(L)]  
    return parameters

def update_parameters_with_momentum(gradient,parameters,LearnRate,v,momentum=0.9):
    #w:=w-lr*dw  ;   b:=b-lr*db
    Layer_num = len(parameters) // 6
    for L in range(1,Layer_num+1): #遍历[1,2,3,4]
        v["dW"+str(L)] = momentum*v["dW"+str(L)] + gradient["dW"+str(L)]   
        parameters["W"+str(L)] = parameters["W"+str(L)] - LearnRate*v["dW"+str(L)]
        v["db"+str(L)] = momentum*v["db"+str(L)] + gradient["db"+str(L)] 
        parameters["b"+str(L)] = parameters["b"+str(L)] - LearnRate*v["db"+str(L)]
    return parameters,v

def update_parameters_with_RMSProp(gradient,parameters,LearnRate,s,decay=0.9,esp=1e-8):
    #w:=w-lr*dw  ;   b:=b-lr*db
    Layer_num = len(parameters) // 6
    for L in range(1,Layer_num+1): #遍历[1,2,3,4]
        s["dW"+str(L)] = decay*s["dW"+str(L)] + (1-decay)*gradient["dW"+str(L)]**2  
        parameters["W"+str(L)] = parameters["W"+str(L)] - LearnRate*gradient["dW"+str(L)]/np.sqrt(s["dW"+str(L)]+esp)
        s["db"+str(L)] = decay*s["db"+str(L)] + (1-decay)*gradient["db"+str(L)]**2 
        parameters["b"+str(L)] = parameters["b"+str(L)] - LearnRate*gradient["db"+str(L)]/np.sqrt(s["db"+str(L)]+esp)
    return parameters,s


def cut_data(train_set_x,train_set_y,batch_size=64):
    m = train_set_x.shape[1]
    #1.打乱数据集
    permutation=list(np.random.permutation(m))
    X_shuffled = train_set_x[:,permutation]
    Y_shuffled = train_set_y[:,permutation]
    
    train_set_cutted=[]
    #2.分批
    num_batch = m//batch_size  #3
    
    for i in range(num_batch):
        mini_batch_X = X_shuffled[:,i*batch_size:(i+1)*batch_size]
        mini_batch_Y = Y_shuffled[:,i*batch_size:(i+1)*batch_size]
        mini_batch = (mini_batch_X,mini_batch_Y)
        train_set_cutted.append(mini_batch)
        
    if m%batch_size!=0:
        mini_batch_X = X_shuffled[:,(i+1)*batch_size:]
        mini_batch_Y = Y_shuffled[:,(i+1)*batch_size:]
        mini_batch = (mini_batch_X,mini_batch_Y)
        train_set_cutted.append(mini_batch)
        
    return train_set_cutted
            
    
def Train_Net(fc_net,train_set_x,train_set_y,batch_size,check=False,num_epoch=500,
                          LearnRate=0.01,active_func="ReLU",optimizer="SGD",use_bn=False,bn_mode="train"):
    #4.初始化参数
    if(optimizer=="momentum"):    
        parameters,v = init_parameters_momentum(fc_net)
    elif(optimizer=="RMSProp"):
        parameters,s = init_parameters_RMSProp(fc_net)
    else:
        parameters = init_parameters(fc_net)
    costs = []
    for epoch in range(0,num_epoch):
        train_set_cutted = cut_data(train_set_x,train_set_y,batch_size)
        for minibatch in train_set_cutted:#遍历所有批次，每一个批次都更新一次参数，从而加快收敛速度
            mini_batch_X,mini_batch_Y = minibatch
            AL,cache,bn_cache,parameters= forward_pass(mini_batch_X,parameters,bn_mode,use_bn,active_func) 
            #6.计算代价值
            cost = compute_cost(AL,mini_batch_Y)         
            if epoch%500 == 0: 
                print("epoch = ",epoch,";    cost = ",cost)
                costs.append(cost)
            #7.反向传播计算梯度
            gradient = backward_pass(AL,parameters,cache,bn_cache,mini_batch_Y,use_bn,active_func)
            #8.根据梯度更新一次参数
            if(optimizer=="momentum"):    
                parameters,v = update_parameters_with_momentum(gradient,parameters,LearnRate,v)
            elif(optimizer=="RMSProp"):
                parameters,s = update_parameters_with_RMSProp(gradient,parameters,LearnRate,s)
            else:
                parameters = update_parameters(gradient,parameters,LearnRate,use_bn)

    plt.plot(costs,'r')
    plt.xlabel("iteration")
    plt.ylabel("cost")
    plt.show()
    return parameters

def Predict(A0,Y,parameters,active_func,bn_mode="test",use_bn=True):
    AL,_0,_1,_2 = forward_pass(A0,parameters,bn_mode,use_bn,active_func)
    m = AL.shape[1]
    p = np.zeros(AL.shape)
    for i in range(0,AL.shape[1]):
        if AL[0,i]>0.5:
            p[0,i]=1
        else:
            p[0,i]=0           
    accuracy = (1/m)* np.sum(p==Y)
    print("accuracy =",accuracy)
            
    
if __name__ == '__main__':
    #1.加载数据
    train_set_x,train_set_y,test_set_x,test_set_y = load_dataset()
    #2.对输入像素值做归一化（0~255）->(0~1)
    train_set_x = train_set_x/255.
    test_set_x = test_set_x/255.
    #3.定义全连接神经网络各层神经元个数，并初始化参数w和b
    #fc_net = [12288,10,3,2,1]  
    fc_net = [12288,20,20,20,20,1] 
    #4.开始训练   #考虑到计算机硬件结构，一般把batch_size设置为16、32、64、128、256，可以加快计算速度  #False True
    parameters = Train_Net(fc_net,train_set_x,train_set_y,batch_size=64,num_epoch=6001,LearnRate=0.01,
                               active_func="tanh",optimizer="SGD",use_bn=True) 
    #5.开始预测
    print("===predict train_set===")
    Predict(train_set_x,train_set_y,parameters,active_func="tanh",use_bn=True)
    print("===predict test_set===")
    Predict(test_set_x,test_set_y,parameters,active_func="tanh",use_bn=True)
    
    

    

    



    


"""Contains the base Layer class, from which all layers inherit.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy as np


class Pooling():
    @staticmethod
    def init(layer, input_shape):
        pass
    @staticmethod
    def forword(layer, flow_data):
        # Z = np.dot(layer['weight'], flow_data) + layer['bias']
        return flow_data
    @staticmethod
    def backword(flow_data, layer, config):
        return flow_data
        # layer["weight_gradient"] = (1 / config['batch_size']) * np.dot(flow_data, layer['input'].T)
        # layer["bias_gradient"] = (1 / config['batch_size']) * np.sum(flow_data, axis=1, keepdims=True)
        # flow_data = np.dot(layer['weight'].T, flow_data)
        # return layer, flow_data



"""Training-related part of the Keras engine.
"""
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
#
# import warnings
# import copy
import numpy as np
from .layer import *
import matplotlib.pyplot as plt


class AADeepLearning():
    solver = None
    net = None
    # 初始化网络参数
    net_parameters = {}
    # 损失值
    loss = []
    # 训练数据 shape: (60000, 28, 28, 1)  (样本数, 宽, 高, 通道数)
    train_data = []
    # 训练数据标签
    train_label = []
    # 损失值
    test_data = []
    # 损失值
    test_lable = []
    # 损失值
    input_shape = 0
    # 学习率
    learning_rate = 0
    # 神经网络层数
    layer_number = 0
    # 神经网络参数 weight和bias
    neural_net = {}

    def __init__(self, solver, net):
        self.solver = solver
        self.net = net
        # 网络结构和定义层一致
        self.net_parameters = net
        self.learning_rate = solver['learning_rate']

    def train(self, train_data, train_label, test_data, test_label):
        # todo 数据量太大会占用太多内存
        self.train_data = train_data
        self.train_label = train_label
        self.test_data = test_data
        self.test_lable = test_label
        # 宽*高*通道数
        self.input_shape = self.train_data.shape[1] * self.train_data.shape[2] * self.train_data.shape[3]
        # 1，初始化网络参数
        neural_net = self.init_parameters()
        loss_list = []
        for iteration in range(self.solver['number_iteration']):
            train_data_batch, train_label_batch = self.next_batch(train_data, train_label, self.solver['batch_size'])
            # print
            # 2，前向传播
            layer_output, cache = self.forward_pass(neural_net, train_data_batch)
            # 3，计算损失
            loss = self.compute_cost(layer_output, train_label_batch)
            loss_list.append(loss)
            # 4，反向传播，求梯度
            gradient = self.backward_pass(neural_net, layer_output, cache, train_label_batch)
            # gradient_check()
            # 5，根据梯度更新一次参数
            neural_net = self.update_parameters(gradient, neural_net)
            if iteration % 100 == 0:
                # self.check_weight(gradient, cache)
                print('iteration:', iteration, 'loss:', loss)
        self.neural_net = neural_net
        # plt.plot(loss_list, 'r')
        # plt.xlabel("iteration")
        # plt.ylabel("loss")
        # plt.show()

    def init_parameters(self):
        # 上一层的输入尺寸
        front_layer_output_shape = self.input_shape
        neural_net = {}
        i = 0
        for key, value in enumerate(self.net_parameters):
            if value['type'] == 'fully_connected':
                i += 1
                # 何凯明初始化，主要针对relu激活函数
                if "weight_init" in value.keys() and value["weight_init"] == 'msra':
                    neural_net["weight_" + str(i)] = np.random.randn(value['neurons_number'],
                                                                 front_layer_output_shape) * (np.sqrt(2/front_layer_output_shape))
                # xavier，主要针对tanh激活函数
                elif "weight_init" in value.keys() and value["weight_init"] == 'xavier':
                    neural_net["weight_" + str(i)] = np.random.randn(value['neurons_number'],
                                                                 front_layer_output_shape) * (np.sqrt(1/front_layer_output_shape))
                else:
                    neural_net["weight_" + str(i)] = np.random.randn(value['neurons_number'],
                                                                 front_layer_output_shape) * 0.01
                neural_net["bias_" + str(i)] = np.zeros((value['neurons_number'], 1))
                # 权重
                # self.net_parameters[key]['weight'] = np.random.randn(value['neurons_number'],front_layer_output_shape)
                # 偏置参数
                # self.net_parameters[key]['bias'] = np.ones((value['neurons_number'], 1))
                front_layer_output_shape = value['neurons_number']
            elif value['type'] == 'relu' or value['type'] == 'sigmoid' or value['type'] == 'tanh':
                neural_net["active_" + str(i)] = value['type']
            elif value['type'] == 'softmax':
                neural_net["output_" + str(i)] = value['type']
        self.layer_number = i
        for i in range(1, self.layer_number + 1):
            print("weight_" + str(i) + ".shape：", neural_net["weight_" + str(i)].shape)
            print("bias_" + str(i) + ".shape：", neural_net["bias_" + str(i)].shape)
        return neural_net

    def forward_pass(self, neural_net, data):
        # X_train shape: (60000, 28, 28, 1) ——> (784, 60000)
        data = data.reshape(data.shape[0], -1).T
        A = data
        cache = {}
        cache["A_0"] = A
        for i in range(1, self.layer_number + 1):
            Z = np.dot(neural_net['weight_' + str(i)], A) + neural_net['bias_' + str(i)]
            if "active_" + str(i) in neural_net.keys() and neural_net["active_" + str(i)] == 'relu':
                A = self.ReLU(Z)
            elif "active_" + str(i) in neural_net.keys() and neural_net["active_" + str(i)] == 'sigmoid':
                A = self.sigmoid(Z)
            elif "active_" + str(i) in neural_net.keys() and neural_net["active_" + str(i)] == 'tanh':
                A = self.tanh(Z)
            else:
                # 没有激活函数
                A = Z

            cache["Z_" + str(i)] = Z
            cache["A_" + str(i)] = A

            if "output_" + str(i) in neural_net.keys() and neural_net["output_" + str(i)] == 'softmax':
                for j in range(self.solver['batch_size']):
                    A[:, j] = np.exp(A[:, j]) / np.sum(np.exp(A[:, j]))

            # print("第"+str(i)+"层，反向传播输出数据尺寸：", Z.shape,"第"+str(i)+"层，前向传播输出数据尺寸：",  cache["A_" + str(i)].shape)
        # Z = np.dot(neural_net['weight_' + str(self.layer_number)], A) + neural_net['bias_' + str(self.layer_number)]
        # A = self.ReLU(Z)
        # cache["Z_"+str(self.layer_number)] = Z
        # cache["A_"+str(self.layer_number)] = A
        # 最后的softmax
        return A, cache

    def compute_cost(self, layer_output, batch_label):
        # print(layer_output.shape)
        # print(batch_label.shape)
        # exit()
        batch_size = self.solver['batch_size']
        loss = 0.0
        for i in range(batch_size):
            loss += -np.sum(np.dot(batch_label[i], np.log(layer_output[:, i])))
        loss = loss / batch_size
        return loss

    def backward_pass(self, neural_net, layer_output, cache, train_label):
        """
        反向传播
        :param layer_output: 前向传播的输出
        :return:
        """
        gradient = {}  # 保存各层参数梯度的字典
        # 获取最末层误差信号 softmax反向传播
        dZL = layer_output - train_label.T
        gradient["d_weight_" + str(self.layer_number)] = (1 / self.solver['batch_size']) * np.dot(dZL, cache[
            "A_" + str(self.layer_number - 1)].T)
        gradient["d_bias_" + str(self.layer_number)] = (1 / self.solver['batch_size']) * np.sum(dZL, axis=1,
                                                                                                keepdims=True)

        for i in reversed(range(1, self.layer_number)):

            dZL = np.dot(neural_net['weight_' + str(i + 1)].T, dZL)

            print("第" + str(i) + "层，反向传播输出数据尺寸：", dZL.shape, "第" + str(i) + "层，前向传播输出数据尺寸：", cache["A_" + str(i)].shape)
            if "active_" + str(i) in neural_net.keys() and neural_net["active_" + str(i)] == 'relu':
                # print(i, dZL.shape, cache["A_" + str(i)].shape)
                # drelu/dz  = 小于零就始终等于0 ，大于0就等于一
                dZL = dZL * np.array(cache["A_" + str(i)] > 0)
            elif "active_" + str(i) in neural_net.keys() and neural_net["active_" + str(i)] == 'sigmoid':
                # dsigmoid/dz  = a*(1-a)
                dZL = dZL * (cache["A_" + str(i)] * (1 - cache["A_" + str(i)]))
            elif "active_" + str(i) in neural_net.keys() and neural_net["active_" + str(i)] == 'tanh':
                # dtanh/dz  = 1-a^2
                dZL = dZL * (1 - np.power(cache["A_" + str(i)], 2))

            gradient["d_weight_" + str(i)] = (1 / self.solver['batch_size']) * np.dot(dZL, cache["A_" + str(i - 1)].T)
            gradient["d_bias_" + str(i)] = (1 / self.solver['batch_size']) * np.sum(dZL, axis=1, keepdims=True)
        return gradient

    def update_parameters(self, gradient, neural_net):
        for index in range(1, self.layer_number + 1):
            neural_net['weight_' + str(index)] = neural_net['weight_' + str(index)] - self.learning_rate * gradient[
                'd_weight_' + str(index)]
            neural_net['bias_' + str(index)] = neural_net['bias_' + str(index)] - self.learning_rate * gradient[
                'd_bias_' + str(index)]
        return neural_net

    # def evaluate(self, test_data, test_label):
    #     print("evaluate")

    def predict(self, test_data, test_label):
        forward_output, _ = self.forward_pass(self.neural_net, test_data)
        forward_output = np.array(forward_output).T
        # print(forward_output.T.shape)
        # print(forward_output.T[0])
        batch_size = forward_output.shape[0]
        # p = np.zeros(test_label.shape)
        right = 0
        for i in range(0, batch_size):
            index = np.argmax(forward_output[i])
            if test_label[i][index] == 1:
                right += 1
        accuracy = right / batch_size
        print(accuracy)
        return accuracy

    def flatten_layer(self):
        self.train_data = []

    def ReLU(self, input):
        return np.maximum(0, input)

    def sigmoid(self, input):
        return 1 / (1 + np.exp(-input))

    def tanh(self, Z):
        return np.tanh(Z)

    def next_batch(self, train_data, train_label, batch_size):
        index = [i for i in range(0, len(train_label))]
        np.random.shuffle(index)
        batch_data = []
        batch_label = []
        for i in range(0, batch_size):
            batch_data.append(train_data[index[i]])
            batch_label.append(train_label[index[i]])
        batch_data = np.array(batch_data)
        batch_label = np.array(batch_label)
        return batch_data, batch_label
    def check_weight(self, gradient, cache):
        for i in range(1, self.layer_number+1):
            print("Z_"+str(i)+"[:,1]_mean =", np.mean(cache["Z_"+str(i)][:, 1]), "; Z1[:,1]_std =",
                  np.std(cache["Z_"+str(i)][:, 1]))  # 统计某一列激活值的均值和标准差
            print("A"+str(i)+"[:,1]_mean =", np.mean(cache["A_"+str(i)][:, 1]), "; A1[:,1]_std =",
                  np.std(cache["A_"+str(i)][:, 1]))  # 统计某一列激活值的均值和标准差
            print("|dW"+str(i)+"|<1e-8 :", np.sum(abs(gradient["d_weight_"+str(i)]) < 1e-8), "/",
                  gradient["d_weight_"+str(i)].shape[0] * gradient["d_weight_"+str(i)].shape[1])



"""Contains the base Layer class, from which all layers inherit.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy as np


class SoftMax():
    @staticmethod
    def init(layer):
        return layer
    @staticmethod
    def forword(flow_data, config):
        for i in range(config['batch_size']):
            flow_data[:, i] = np.exp(flow_data[:, i]) / np.sum(np.exp(flow_data[:, i]))
        return flow_data
    @staticmethod
    def backword(flow_data, label):
        # 获取最末层误差信号 softmax反向传播
        return flow_data - label.T



"""Contains the base Layer class, from which all layers inherit.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import copy
import re
from six.moves import zip

# from .. import backend as K
# from .. import initializers


class Layer(object):

    def __init__(self, **kwargs):
        self.input_spec = None

    @staticmethod
    def _node_key(layer, node_index):
        """Converts a layer and its index to a unique (immutable type) name.

        This function is used internally with `self._network_nodes`.

        # Arguments
            layer: The layer.
            node_index: The layer's position (e.g. via enumerate) in a list of
                nodes.

        # Returns
            The unique name.
        """
        return layer.name + '_ib-' + str(node_index)

    @property
    def losses(self):
        return self._losses

    @property
    def updates(self):
        if not self.trainable and not self.stateful:
            return []
        return self._updates

    @property
    def built(self):
        return self._built

    @built.setter
    def built(self, value):
        self._built = value

    @property
    def trainable_weights(self):
        trainable = getattr(self, 'trainable', True)
        if trainable:
            return self._trainable_weights
        else:
            return []

    @trainable_weights.setter
    def trainable_weights(self, weights):
        self._trainable_weights = weights



"""Contains the base Layer class, from which all layers inherit.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy as np


class Convolutional():
    @staticmethod
    def init(layer, input_shape):
        # 何凯明初始化，主要针对relu激活函数
        # if layer["weight_init"] == 'msra':
        #     layer["weight"] = np.random.randn(layer['neurons_number'],
        #                                       input_shape) * (np.sqrt(2/input_shape))
        # # xavier，主要针对tanh激活函数
        # elif layer["weight_init"] == 'xavier':
        #     layer["weight"] = np.random.randn(layer['neurons_number'],
        #                                       input_shape) * (np.sqrt(1/input_shape))
        # else:
        #     layer["weight"] = np.random.randn(layer['neurons_number'], input_shape) * 0.01
        layer["weight"] = np.random.randn(layer['kernel_width'], layer['kernel_height']) * 0.01
        layer["bias"] = np.zeros((layer['kernel_width'], layer['kernel_height']))
        return layer, input_shape
    @staticmethod
    def forword(layer, flow_data):
        # Z = np.dot(layer['weight'], flow_data) + layer['bias']
        return flow_data
    @staticmethod
    def backword(flow_data, layer, config):
        # layer["weight_gradient"] = (1 / config['batch_size']) * np.dot(flow_data, layer['input'].T)
        # layer["bias_gradient"] = (1 / config['batch_size']) * np.sum(flow_data, axis=1, keepdims=True)
        # flow_data = np.dot(layer['weight'].T, flow_data)
        return layer, flow_data



"""Training-related part of the Keras engine.
"""
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
#
# import warnings
# import copy
import numpy as np
from .layer.fully_connected import FullyConnected
from .layer.relu import Relu
from .layer.softmax import SoftMax
from .layer.convolutional import Convolutional
from .layer.pooling import Pooling
import matplotlib.pyplot as plt

"""
如果每一层神经元的尺寸如下：
weight_1.shape： (256, 784)
bias_1.shape： (256, 1)
weight_2.shape： (128, 256)
bias_2.shape： (128, 1)
weight_3.shape： (10, 128)
bias_3.shape： (10, 1)

那么数据前向传播和反向传播尺寸的变化情况如下

第1层，前向传播输出数据尺寸： (256, 64)  =    神经元尺寸 (256, 784) *数据尺寸(784,64)
第2层，前向传播输出数据尺寸： (128, 64)  =     神经元尺寸 (128, 256)*数据尺寸(256, 64)
第3层，前向传播输出数据尺寸： (10, 64)    =      神经元尺寸  (10, 128) *数据尺寸(128, 64)

第3层，因为是从softmax层回传过来的，尺寸不变 (10, 64)
第2层，反向传播输出数据尺寸： (128, 64)  =   下一层weight 尺寸.T (128, 10)*流动数据尺寸(10, 64)
第1层，反向传播输出数据尺寸： (256, 64)  =   下一层weight尺寸.T (256, 128)*流动数据尺寸(128, 64)
"""

class AADeepLearning():
    config = None
    net = None
    # 初始化网络参数
    net = {}
    # 损失值
    loss = []
    # 训练数据 shape: (60000, 28, 28, 1)  (样本数, 宽, 高, 通道数)
    train_data = []
    # 训练数据标签
    train_label = []
    # 损失值
    test_data = []
    # 损失值
    test_lable = []
    # 损失值
    input_shape = 0
    # 学习率
    learning_rate = 0
    # 神经网络层数
    layer_number = 0
    # 神经网络参数 weight和bias
    net = {}

    def __init__(self, config, net):
        self.config = config
        self.net = net
        # 网络结构和定义层一致
        self.net = net
        self.learning_rate = config['learning_rate']

    def train(self, train_data, train_label, test_data, test_label):
        # 宽*高*通道数
        self.input_shape = train_data.shape[1] * train_data.shape[2] * train_data.shape[3]
        # 1，初始化网络参数
        net = self.init_parameters()
        loss_list = []
        for iteration in range(self.config['number_iteration']):
            train_data_batch, train_label_batch = self.next_batch(train_data, train_label, self.config['batch_size'])
            # 2，前向传播
            flow_data = self.forward_pass(net, train_data_batch)
            # 3，计算损失
            loss = self.compute_cost(flow_data, train_label_batch)
            loss_list.append(loss)
            # 4，反向传播，求梯度
            net = self.backward_pass(net, flow_data, train_label_batch)
            # 5，根据梯度更新一次参数
            net = self.update_parameters(net)
            if iteration % 100 == 0:
                # self.check_weight(gradient, cache)
                print('iteration:', iteration, 'loss:', loss)
        self.net = net
        # plt.plot(loss_list, 'r')
        # plt.xlabel("iteration")
        # plt.ylabel("loss")
        # plt.show()

    def init_parameters(self):
        net = self.net
        # 第一层的输出尺寸就是数据尺寸
        output_shape=self.input_shape
        for i, layer in enumerate(net):
            if layer['type'] == 'convolutional':
                net[i], output_shape = Convolutional.init(layer=layer, input_shape=output_shape)
            elif layer['type'] == 'fully_connected':
                net[i], output_shape = FullyConnected.init(layer=layer, input_shape=output_shape)
        for i, layer in enumerate(net):
            if layer['type'] == 'fully_connected':
                print(layer['name']+",weight.shape：", layer["weight"].shape)
                print(layer['name']+",bias.shape：", layer["bias"].shape)
        return net

    def forward_pass(self, net, data):
        # X_train shape: (60000, 28, 28, 1) ——> (784, 60000)
        # 流动数据，一层一层的计算，并先后流动
        flow_data = data.reshape(data.shape[0], -1).T
        for i, layer in enumerate(net):
            # 缓存当前层的输入
            net[i]['input'] = flow_data
            if layer['type'] == 'convolutional':
                flow_data = Convolutional.forword(layer=layer, flow_data=flow_data)
            elif layer['type'] == 'pooling':
                flow_data = Pooling.forword(layer=layer, flow_data=flow_data)
            elif layer['type'] == 'fully_connected':
                flow_data = FullyConnected.forword(layer=layer, flow_data=flow_data)
            elif layer['type'] == 'relu':
                flow_data = Relu.forword(flow_data=flow_data)
            elif layer['type'] == 'softmax':
                flow_data = SoftMax.forword(flow_data=flow_data, config=self.config)
            # 缓存当前层的输出
            net[i]['output'] = flow_data
        return flow_data

    def compute_cost(self, layer_output, batch_label):
        batch_size = self.config['batch_size']
        loss = 0.0
        for i in range(batch_size):
            loss += -np.sum(np.dot(batch_label[i], np.log(layer_output[:, i])))
        loss = loss / batch_size
        return loss

    def backward_pass(self, net, flow_data, train_label):
        layer_number = len(net)
        for i in reversed(range(0, layer_number)):
            layer = net[i]
            if layer['type'] == 'convolutional':
                net[i], flow_data = Convolutional.backword(flow_data=flow_data, layer=layer, config=self.config)
            elif layer['type'] == 'pooling':
                flow_data = Pooling.backword(flow_data=flow_data, layer=layer, config=self.config)
            elif layer['type'] == 'fully_connected':
                net[i], flow_data = FullyConnected.backword(flow_data=flow_data, layer=layer, config=self.config)
            elif layer['type'] == 'relu':
                # print(layer["name"], flow_data.shape, layer["output"].shape)
                # exit()
                flow_data = Relu.backword(flow_data=flow_data, layer=layer)
            elif layer['type'] == 'softmax':
                flow_data = SoftMax.backword(flow_data=flow_data, label=train_label)
        return net

    def update_parameters(self, net):

        for i, layer in enumerate(net):
            # if layer['type'] == 'fully_connected' or layer['type'] == 'convolutional':
            if layer['type'] == 'fully_connected':
                net[i]["weight"] = net[i]["weight"] - self.learning_rate * net[i]["weight_gradient"]
                net[i]["bias"] = net[i]["bias"] - self.learning_rate * net[i]["bias_gradient"]
        return net

    def predict(self, test_data, test_label):
        flow_data = self.forward_pass(self.net, test_data)
        flow_data = np.array(flow_data).T
        batch_size = flow_data.shape[0]
        right = 0
        for i in range(0, batch_size):
            index = np.argmax(flow_data[i])
            if test_label[i][index] == 1:
                right += 1
        accuracy = right / batch_size
        print(accuracy)
        return accuracy

    def flatten_layer(self):
        self.train_data = []

    def ReLU(self, input):
        return np.maximum(0, input)

    def sigmoid(self, input):
        return 1 / (1 + np.exp(-input))

    def tanh(self, Z):
        return np.tanh(Z)

    def next_batch(self, train_data, train_label, batch_size):
        index = [i for i in range(0, len(train_label))]
        np.random.shuffle(index)
        batch_data = []
        batch_label = []
        for i in range(0, batch_size):
            batch_data.append(train_data[index[i]])
            batch_label.append(train_label[index[i]])
        batch_data = np.array(batch_data)
        batch_label = np.array(batch_label)
        return batch_data, batch_label
    def check_weight(self, gradient, cache):
        for i in range(1, self.layer_number+1):
            print("Z_"+str(i)+"[:,1]_mean =", np.mean(cache["Z_"+str(i)][:, 1]), "; Z1[:,1]_std =",
                  np.std(cache["Z_"+str(i)][:, 1]))  # 统计某一列激活值的均值和标准差
            print("A"+str(i)+"[:,1]_mean =", np.mean(cache["A_"+str(i)][:, 1]), "; A1[:,1]_std =",
                  np.std(cache["A_"+str(i)][:, 1]))  # 统计某一列激活值的均值和标准差
            print("|dW"+str(i)+"|<1e-8 :", np.sum(abs(gradient["d_weight_"+str(i)]) < 1e-8), "/",
                  gradient["d_weight_"+str(i)].shape[0] * gradient["d_weight_"+str(i)].shape[1])



"""Contains the base Layer class, from which all layers inherit.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy as np


class FullyConnected():
    @staticmethod
    def init(layer, input_shape):
        # 何凯明初始化，主要针对relu激活函数
        if layer["weight_init"] == 'msra':
            layer["weight"] = np.random.randn(layer['neurons_number'],
                                              input_shape) * (np.sqrt(2/input_shape))
        # xavier，主要针对tanh激活函数
        elif layer["weight_init"] == 'xavier':
            layer["weight"] = np.random.randn(layer['neurons_number'],
                                              input_shape) * (np.sqrt(1/input_shape))
        else:
            layer["weight"] = np.random.randn(layer['neurons_number'], input_shape) * 0.01
        layer["bias"] = np.zeros((layer['neurons_number'], 1))
        return layer, layer['neurons_number']
    @staticmethod
    def forword(layer, flow_data):
        Z = np.dot(layer['weight'], flow_data) + layer['bias']
        return Z
    @staticmethod
    def backword(flow_data, layer, config):
        layer["weight_gradient"] = (1 / config['batch_size']) * np.dot(flow_data, layer['input'].T)
        layer["bias_gradient"] = (1 / config['batch_size']) * np.sum(flow_data, axis=1, keepdims=True)
        flow_data = np.dot(layer['weight'].T, flow_data)
        return layer, flow_data



#!/usr/bin/env python
#-*- coding:utf-8 -*-

#############################################
# File Name: setup.py
# Author: mage
# Mail: mage@woodcol.com
# Created Time:  2018-1-23 19:17:34
#############################################


from setuptools import setup, find_packages

setup(
    name = "aadeeplearning",
    version = "1.0.1",
    keywords = ("pip", "aadeeplearning"),
    description = "aa deep learning",
    long_description = "aa deep learning",
    license = "MIT Licence",

    url = "https://github.com/luojiangtao/aadeeplearning",
    author = "luojiangtao",
    author_email = "1368761119@qq.com",

    packages = find_packages(),
    include_package_data = True,
    platforms = "any",
    install_requires = []
)


import h5py
import matplotlib.pyplot as plt
import numpy as np


# 字典！以键值对的方式保存的一种数据结构
def load_dataset():
    # 1.创建文件对象
    train_dataset = h5py.File('./我的数据集/train_catvnoncat.h5', 'r')
    test_dataset = h5py.File('./我的数据集/test_catvnoncat.h5', 'r')
    # 2.读取数据
    train_set_x = np.array(train_dataset["train_set_x"][:])
    train_set_y = np.array(train_dataset["train_set_y"][:])
    test_set_x = np.array(test_dataset["test_set_x"][:])
    test_set_y = np.array(test_dataset["test_set_y"][:])
    plt.figure(figsize=(2, 2))
    plt.imshow(train_set_x[110])
    plt.show()
    # 3.变化维度以适应神经网络输入
    # (209, 64, 64, 3) -> (12288,209)
    train_set_x = train_set_x.reshape(train_set_x.shape[0], -1).T  # (12288,209)
    test_set_x = test_set_x.reshape(test_set_x.shape[0], -1).T  # (12288,209)
    # (209,)->(1,209)
    train_set_y = train_set_y.reshape(1, train_set_y.shape[0])  # (1,209)
    test_set_y = test_set_y.reshape(1, test_set_y.shape[0])  # (1,209)

    return train_set_x, train_set_y, test_set_x, test_set_y


def init_parameters(fc_net):
    # 1.定义一个字典，存放参数矩阵W1，b1，W2，b2，W3，b3，W4，b4
    parameters = {}
    Layer_num = len(fc_net)  # Layer_num=5

    for L in range(1, Layer_num):
        # parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*0.01  #
        # parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*np.sqrt(1/fc_net[L-1]) #Xavier初始化，针对tanh函数
        parameters["W" + str(L)] = np.random.randn(fc_net[L], fc_net[L - 1]) * np.sqrt(
            2 / fc_net[L - 1])  # He初始化（MSRA），针对ReLU函数
        parameters["b" + str(L)] = np.zeros((fc_net[L], 1))

    for L in range(1, Layer_num):
        print("W" + str(L) + "=", parameters["W" + str(L)].shape)
        print("b" + str(L) + "=", parameters["b" + str(L)].shape)

    return parameters


def ReLU(Z):
    return np.maximum(0, Z)


def tanh(Z):
    return np.tanh(Z)


def sigmoid(Z):
    return 1 / (1 + np.exp(-Z))


def forward_pass(A0, parameters, active_func="ReLU"):  # 前向计算函数
    cache = {}
    A = A0
    cache["A0"] = A0
    Layer_num = len(parameters) // 2  # 写成/2代表精确除法，得到一个浮点数，写成//2代表向下取整除法，得到一个整数
    for L in range(1, Layer_num):  # 遍历[1,2,3]
        # z=wx+b
        Z = np.dot(parameters["W" + str(L)], A) + parameters["b" + str(L)]  # python的广播机制
        if active_func == "sigmoid":
            A = sigmoid(Z)
        elif active_func == "tanh":
            A = tanh(Z)
        else:
            A = ReLU(Z)
        cache["A" + str(L)] = A
        cache["Z" + str(L)] = Z

    Z = np.dot(parameters["W" + str(Layer_num)], A) + parameters["b" + str(Layer_num)]  # python的广播机制
    A = sigmoid(Z)
    cache["A" + str(Layer_num)] = A
    cache["Z" + str(Layer_num)] = Z
    return A, cache


def compute_cost(AL, Y):
    m = Y.shape[1]  # Y =（1,209）
    cost = (1 / m) * np.sum((1 / 2) * (AL - Y) * (AL - Y))  # 代价函数
    return cost


def backward_pass(AL, parameters, cache, Y, active_func="ReLU"):
    m = Y.shape[1]  # 样本总数
    gradient = {}  # 保存各层参数梯度的字典
    Layer_num = len(parameters) // 2
    dZL = (AL - Y) * (AL * (1 - AL))  # 获取最末层误差信号 dZL.shape = (1,209)
    gradient["dW" + str(Layer_num)] = (1 / m) * np.dot(dZL, cache["A" + str(Layer_num - 1)].T)
    gradient["db" + str(Layer_num)] = (1 / m) * np.sum(dZL, axis=1, keepdims=True)
    for L in reversed(range(1, Layer_num)):  # 遍历[3,2,1]，其中reversed函数[1,2,3]颠倒为[3,2,1]
        # dZL = np.dot(parameters["W"+str(L+1)].T,dZL)*(AL*(1-AL))  #有问题
        if active_func == "sigmoid":
            dZL = np.dot(parameters["W" + str(L + 1)].T, dZL) * (
                        cache["A" + str(L)] * (1 - cache["A" + str(L)]))  # dsigmoid/dz  = a*(1-a)
        elif active_func == "tanh":
            dZL = np.dot(parameters["W" + str(L + 1)].T, dZL) * (
                        1 - np.power(cache["A" + str(L)], 2))  # dtanh/dz  = 1-a^2
        else:
            dZL = np.dot(parameters["W" + str(L + 1)].T, dZL) * np.array(cache["Z" + str(L)] > 0)

        gradient["dW" + str(L)] = (1 / m) * np.dot(dZL, cache["A" + str(L - 1)].T)
        gradient["db" + str(L)] = (1 / m) * np.sum(dZL, axis=1, keepdims=True)
    return gradient


def update_parameters(gradient, parameters, LearnRate):
    # w:=w-lr*dw  ;   b:=b-lr*db
    Layer_num = len(parameters) // 2
    for L in range(1, Layer_num + 1):  # 遍历[1,2,3,4]
        parameters["W" + str(L)] = parameters["W" + str(L)] - LearnRate * gradient["dW" + str(L)]
        parameters["b" + str(L)] = parameters["b" + str(L)] - LearnRate * gradient["db" + str(L)]
    return parameters


# 梯度检验
# 解析法：求得梯度解析表达式，通过这个表达式得到梯度（确切解）
# 数值逼近：（近似解）
def grad_dict_to_vector(gradient):  # 梯度字典转列矩阵
    Layer_num = len(gradient) // 2  # gradients=4
    count = 0
    for L in range(1, Layer_num + 1):  # 遍历[1,2,...,Layer_num]
        dW_vector = np.reshape(gradient["dW" + str(L)], (-1, 1))  # 将该层dW矩阵展平为一个列矩阵 （4,12288）->（49152,1）
        db_vector = np.reshape(gradient["db" + str(L)], (-1, 1))  # 将该层db矩阵展平为一个列矩阵 （4,1）->（4,1）
        vec_L = np.concatenate((dW_vector, db_vector), axis=0)  # 先将该层W个b串联叠加
        if count == 0:
            vec_output = vec_L  # 叠加到输出列矩阵
        else:
            vec_output = np.concatenate((vec_output, vec_L), axis=0)  # 逐层串联叠加
        count = count + 1
    return vec_output  # 返回列矩阵


def param_dict_to_vector(parameters):  # 参数字典转列矩阵
    # 将我们所有的参数字典转换为满足我们特定所需形状的单个向量。
    Layer_num = len(parameters) // 2  # Layer_num=4
    count = 0
    for L in range(1, Layer_num + 1):  # 遍历[1,2,...,Layer_num]
        W_vector = np.reshape(parameters["W" + str(L)], (-1, 1))  # 将该层W参数矩阵展平为一个列矩阵
        b_vector = np.reshape(parameters["b" + str(L)], (-1, 1))  # 将该层b参数矩阵展平为一个列矩阵
        vec_L = np.concatenate((W_vector, b_vector), axis=0)  # 串联叠加
        if count == 0:
            vec_output = vec_L
        else:
            vec_output = np.concatenate((vec_output, vec_L), axis=0)  # 串联叠加
        count = count + 1
    return vec_output


def vector_to_param_dict(vec, param_src):  # 列矩阵转参数字典,第一个输入为列矩阵，第二个输入为保存W和b的参数字典
    Layer_num = len(param_src) // 2  # Layer_num=4
    param_epsilon = param_src
    idx_start = 0
    idx_end = 0
    for L in range(1, Layer_num + 1):  # 遍历[1,2,...,Layer_num]
        row = param_src["W" + str(L)].shape[0]
        col = param_src["W" + str(L)].shape[1]
        idx_end = idx_start + row * col  # 该W参数矩阵元素个数
        # print("idx_start=",idx_start,";idx_end=",idx_end)
        param_epsilon["W" + str(L)] = vec[idx_start:idx_end].reshape((row, col))
        idx_start = idx_end

        row = param_src["b" + str(L)].shape[0]
        col = param_src["b" + str(L)].shape[1]
        idx_end = idx_start + row * col  # 该b参数矩阵元素个数
        # print("b.shape=",param_src["b"+str(L)].shape)
        # print("idx_start=",idx_start,";idx_end=",idx_end)
        param_epsilon["b" + str(L)] = vec[idx_start:idx_end].reshape((row, col))
        idx_start = idx_end
    return param_epsilon  # 返回添加了epsilon的参数字典


def gradient_check(A0, Y, gradient, parameters, epsilon=1e-4):
    grad_vec = grad_dict_to_vector(gradient)  # 梯度字典转列向量
    param_vec = param_dict_to_vector(parameters)  # 参数字典转列向量(49182,1)
    param_num = param_vec.shape[0]  # 49182
    grad_vec_approach = np.zeros(grad_vec.shape)
    for i in range(param_num):
        if i % 1000 == 0:
            print("grad checking i=", i)
        param_vec_plus = np.copy(param_vec)
        param_vec_plus[i][0] = param_vec_plus[i][0] + epsilon
        AL, _ = forward_pass(A0, vector_to_param_dict(param_vec_plus, parameters))
        J_plus_epsilon = compute_cost(AL, Y)

        param_vec_minus = np.copy(param_vec)
        param_vec_minus[i][0] = param_vec_minus[i][0] - epsilon
        AL, _ = forward_pass(A0, vector_to_param_dict(param_vec_minus, parameters))
        J_minus_epsilon = compute_cost(AL, Y)

        grad_vec_approach[i][0] = (J_plus_epsilon - J_minus_epsilon) / (2 * epsilon)

    # 在机器学习中，表征两个向量之间差异性的方法：L2范数（欧氏距离）、余弦距离
    # L2范数：主要用于表征两个向量之间数值的差异（适合我们现在的情况）
    # 余弦距离：主要用于表征两个向量之间方向的差异
    diff = np.sqrt(np.sum((grad_vec - grad_vec_approach) ** 2)) / (
                np.sqrt(np.sum((grad_vec) ** 2)) + np.sqrt(np.sum((grad_vec_approach) ** 2)))
    if diff > 1e-4:
        print("Maybe a mistake in your bakeward pass!!!  diff=", diff)
    else:
        print("No problem in your bakeward pass!!!  diff=", diff)

        """
W1= (4, 12288)
b1= (4, 1)
W2= (3, 4)
b2= (3, 1)
W3= (2, 3)
b3= (2, 1)
W4= (1, 2)
b4= (1, 1)   

def gradient_check(A0,Y,gradient,parameters,check_layer,epsilon=1e-4):
    grad_vec = grad_dict_to_vector(gradient)#梯度字典转列向量
    param_vec = param_dict_to_vector(parameters)#参数字典转列向量(49182,1)
    param_num = param_vec.shape[0]  #49182
    #grad_vec_approach = np.zeros(grad_vec.shape)
    #1.根据指定层数，获取对应层数解析梯度片段
    if check_layer==1:
        start=0
        end=49156
    elif check_layer==2:
        start=49156
        end=49171
    elif check_layer==3:
        start=49171
        end=49179
    elif check_layer==4:
        start=49179
        end=49182
    else:
        start=0
        end=49182
    grad_vec_slice = grad_vec[start:end]
    grad_vec_approach = np.zeros(grad_vec_slice.shape)
    #2.求取数值梯度
    for i in range(start,end):
        if i%1000==0:
            print("grad checking i=",i)
        param_vec_plus = np.copy(param_vec)
        param_vec_plus[i][0] = param_vec_plus[i][0] + epsilon
        AL,_ = forward_pass(A0,vector_to_param_dict(param_vec_plus,parameters))
        J_plus_epsilon = compute_cost(AL,Y)

        param_vec_minus = np.copy(param_vec)
        param_vec_minus[i][0] = param_vec_minus[i][0] - epsilon
        AL,_ = forward_pass(A0,vector_to_param_dict(param_vec_minus,parameters))
        J_minus_epsilon = compute_cost(AL,Y)

        grad_vec_approach[i-start][0]= (J_plus_epsilon-J_minus_epsilon)/(2*epsilon)

    #3.求距离
    diff = np.sqrt(np.sum((grad_vec_slice-grad_vec_approach)**2))/(np.sqrt(np.sum((grad_vec_slice)**2))+np.sqrt(np.sum((grad_vec_approach)**2)))
    if diff > 1e-4:
        print("Maybe a mistake in your bakeward pass!!!  diff=",diff)
    else:
        print("No problem in your bakeward pass!!!  diff=",diff)
        """


def Train_Net(fc_net, train_set_x, train_set_y, check=False, iterations=2000, LearnRate=0.01, active_func="ReLU"):
    # 4.初始化参数
    parameters = init_parameters(fc_net)
    costs = []  # 保存我们每次迭代计算得到的代价值
    for iteration in range(0, iterations):
        AL, cache = forward_pass(train_set_x, parameters, active_func)  # AL=(1,209)
        # 6.计算代价值
        cost = compute_cost(AL, train_set_y)

        if iteration % 500 == 0:  # 每500个迭代周期打印一次代价值
            print("iteration = ", iteration, ";    cost = ", cost)
            costs.append(cost)
        # 7.反向传播计算梯度
        gradient = backward_pass(AL, parameters, cache, train_set_y, active_func)
        if check and iteration == 2000:
            gradient_check(train_set_x, train_set_y, gradient, parameters)
        # 8.根据梯度更新一次参数
        parameters = update_parameters(gradient, parameters, LearnRate)

        if iteration > 500 and iteration % 2000 == 0:
            print("Z1[:,1]_mean =", np.mean(cache["Z1"][:, 1]), "; Z1[:,1]_std =",
                  np.std(cache["Z1"][:, 1]))  # 统计某一列激活值的均值和标准差
            print("Z2[:,1]_mean =", np.mean(cache["Z2"][:, 1]), "; Z2[:,1]_std =",
                  np.std(cache["Z2"][:, 1]))  # 统计某一列激活值的均值和标准差
            print("Z3[:,1]_mean =", np.mean(cache["Z3"][:, 1]), "; Z3[:,1]_std =",
                  np.std(cache["Z3"][:, 1]))  # 统计某一列激活值的均值和标准差
            print("Z4[:,1]_mean =", np.mean(cache["Z4"][:, 1]), "; Z4[:,1]_std =",
                  np.std(cache["Z4"][:, 1]))  # 统计某一列激活值的均值和标准差
            print("A1[:,1]_mean =", np.mean(cache["A1"][:, 1]), "; A1[:,1]_std =",
                  np.std(cache["A1"][:, 1]))  # 统计某一列激活值的均值和标准差
            print("A2[:,1]_mean =", np.mean(cache["A2"][:, 1]), "; A2[:,1]_std =",
                  np.std(cache["A2"][:, 1]))  # 统计某一列激活值的均值和标准差
            print("A3[:,1]_mean =", np.mean(cache["A3"][:, 1]), "; A3[:,1]_std =",
                  np.std(cache["A3"][:, 1]))  # 统计某一列激活值的均值和标准差
            print("A4[:,1]_mean =", np.mean(cache["A4"][:, 1]), "; A4[:,1]_std =",
                  np.std(cache["A4"][:, 1]))  # 统计某一列激活值的均值和标准差
            print("|dW4|<1e-8 :", np.sum(abs(gradient["dW4"]) < 1e-8), "/",
                  gradient["dW4"].shape[0] * gradient["dW4"].shape[1])
            print("|dW3|<1e-8 :", np.sum(abs(gradient["dW3"]) < 1e-8), "/",
                  gradient["dW3"].shape[0] * gradient["dW3"].shape[1])
            print("|dW2|<1e-8 :", np.sum(abs(gradient["dW2"]) < 1e-8), "/",
                  gradient["dW2"].shape[0] * gradient["dW2"].shape[1])
            print("|dW1|<1e-8 :", np.sum(abs(gradient["dW1"]) < 1e-8), "/",
                  gradient["dW1"].shape[0] * gradient["dW1"].shape[1])

    plt.plot(costs, 'r')
    plt.xlabel("iteration")
    plt.ylabel("cost")
    plt.show()
    return parameters


def Predict(A0, Y, parameters, active_func):
    AL, _ = forward_pass(A0, parameters, active_func)  # AL是（1,50）
    m = AL.shape[1]
    p = np.zeros(AL.shape)
    for i in range(0, AL.shape[1]):
        if AL[0, i] > 0.5:
            p[0, i] = 1
        else:
            p[0, i] = 0
    accuracy = (1 / m) * np.sum(p == Y)
    print("accuracy =", accuracy)


if __name__ == '__main__':
    # 1.加载数据
    train_set_x, train_set_y, test_set_x, test_set_y = load_dataset()
    # 2.对输入像素值做归一化（0~255）->(0~1)
    train_set_x = train_set_x / 255.
    test_set_x = test_set_x / 255.
    # 3.定义全连接神经网络各层神经元个数，并初始化参数w和b
    fc_net = [12288, 20, 20, 20, 20, 1]  # 四层网络（梯度消失）
    # fc_net = [784,1000,1000,1000,1000,1000,10]
    # fc_net = [12288,10,3,2,1]       #四层网络（梯度消失）
    # fc_net = [12288,100,1]         #两层网络
    # 4.开始训练
    parameters = Train_Net(fc_net, train_set_x, train_set_y, iterations=3000, LearnRate=0.1, active_func="tanh")
    # 5.开始预测
    Predict(test_set_x, test_set_y, parameters, active_func="tanh")












class Layer(object):
    def __init__(self,name):
        self.name=name

    def forword(self):
        pass
    def backword(self):
        pass
    # @staticmethod
    # def animal_talk(obj): #动物叫的接口
    #     obj.talk()

    # def __getattribute__(self, item):
    #     if item == 'a':
    #         print('------%s------' % item)
    #         return '%s is get' % item
    #     else:
    #         return object.__getattribute__(self, item)
class ALayer(Layer):
    def forword(self):
        print("ALayer forword!")
    def backword(self):
        print("ALayer forword!")
class BLayer(Layer):
    def forword(self):
        print("BLayer forword!")
    def backword(self):
        print("BLayer forword!")

c=ALayer('李丽')
d=BLayer('王张')
list = {"c":c, "d":d}
# Animal.animal_talk(list["c"])#多态：一个接口多种实现
# Animal.animal_talk(list["d"]) #一个接口多种实现

list["c"].forword()
list["d"].forword()


import h5py
import matplotlib.pyplot as plt
import numpy as np

param_num = 2

#字典！以键值对的方式保存的一种数据结构
def load_dataset():
    #1.创建文件对象
    hdf5_path = 'my_train_data/cat_vs_dog/dataset5000.hdf5'
    train_dataset = h5py.File(hdf5_path,'r')
    test_dataset = h5py.File(hdf5_path,'r')
    #2.读取数据
    train_set_x = np.array(train_dataset["train_img"][:])
    train_set_y = np.array(train_dataset["train_labels"][:])
    test_set_x = np.array(test_dataset["val_img"][:])
    test_set_y = np.array(test_dataset["val_labels"][:])
    plt.figure(figsize=(2,2))
    plt.imshow(train_set_x[610])
    #3.变化维度以适应神经网络输入
    #(209, 64, 64, 3) -> (12288,209)
    train_set_x = train_set_x.reshape(train_set_x.shape[0],-1).T  #(12288,209)
    test_set_x = test_set_x.reshape(test_set_x.shape[0],-1).T  #(12288,209)
    #(209,)->(1,209)
    train_set_y = train_set_y.reshape(1,train_set_y.shape[0]) #(1,209)
    test_set_y = test_set_y.reshape(1,test_set_y.shape[0]) #(1,209)
    
    return train_set_x,train_set_y,test_set_x,test_set_y

def init_parameters(fc_net):
    parameters = {}
    Layer_num = len(fc_net) #Layer_num=5  
    for L in range(1,Layer_num):
        #parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*0.01  #
        #parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*np.sqrt(1/fc_net[L-1]) #Xavier初始化，针对tanh函数
        parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*np.sqrt(2/fc_net[L-1]) #He初始化，针对ReLU函数
        parameters["b"+str(L)] = np.zeros((fc_net[L],1))   
        
    for L in range(1,Layer_num):
        print("W"+str(L)+"=",parameters["W"+str(L)].shape)
        print("b"+str(L)+"=",parameters["b"+str(L)].shape)                       
    return parameters

def init_parameters_momentum(fc_net):
    parameters = {}
    v = {}
    Layer_num = len(fc_net)  
    
    for L in range(1,Layer_num):
        #parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*0.01   #最普通的随机初始化
        parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])* np.sqrt(2/fc_net[L-1])
        parameters["b"+str(L)] = np.zeros((fc_net[L],1))  #使用BN的时候其实可以去掉偏置b了（要在很多代码里面改，先不改了......）
        
        v["dW" + str(L)] = np.zeros((fc_net[L],fc_net[L-1]))
        v["db" + str(L)] = np.zeros((fc_net[L],1))

    for L in range(1,Layer_num):
        print("W"+str(L)+"=",parameters["W"+str(L)].shape)
        print("b"+str(L)+"=",parameters["b"+str(L)].shape)                       
    return parameters,v

def ReLU(Z):
    return np.maximum(0,Z)

def tanh(Z):
    return np.tanh(Z)

def sigmoid(Z):
    return 1/(1+np.exp(-Z))

def forward_pass(A0,parameters,active_func="ReLU"):    #前向计算函数
    cache ={}
    A = A0
    cache["A0"] = A0
    Layer_num = len(parameters) // param_num
    for L in range(1,Layer_num): 
        Z = np.dot(parameters["W"+str(L)],A) + parameters["b"+str(L)] 
        if active_func=="sigmoid":
            A = sigmoid(Z)
        elif active_func=="tanh":
            A = tanh(Z)
        else:
            A = ReLU(Z)
        cache["A"+str(L)] = A
        cache["Z"+str(L)] = Z
        
    Z = np.dot(parameters["W"+str(Layer_num)],A) + parameters["b"+str(Layer_num)] #python的广播机制
    A = sigmoid(Z)
    cache["A"+str(Layer_num)] = A
    cache["Z"+str(Layer_num)] = Z
    return A,cache

def compute_cost(AL,Y):
    m = Y.shape[1]   #Y =（1,209）
    cost = (1/m)*np.sum((1/2)*(AL-Y)*(AL-Y)) #代价函数
    return cost

def backward_pass(AL,parameters,cache,Y,active_func="ReLU"):
    m = Y.shape[1]  #样本总数
    gradient = {}   #保存各层参数梯度的字典
    Layer_num = len(parameters) // param_num
    dZL = (AL-Y)*(AL*(1-AL))#获取最末层误差信号 dZL.shape = (1,209)  
    gradient["dW"+str(Layer_num)] = (1/m)*np.dot(dZL,cache["A"+str(Layer_num-1)].T)
    gradient["db"+str(Layer_num)] = (1/m)*np.sum(dZL,axis=1,keepdims = True)    
    for L in reversed(range(1,Layer_num)): #遍历[3,2,1]，其中reversed函数[1,2,3]颠倒为[3,2,1]
        if active_func=="sigmoid":
            dZL = np.dot(parameters["W"+str(L+1)].T,dZL)*(cache["A"+str(L)]*(1-cache["A"+str(L)]))  #dsigmoid/dz  = a*(1-a)
        elif active_func=="tanh":
            dZL = np.dot(parameters["W"+str(L+1)].T,dZL)*(1-np.power(cache["A"+str(L)],2))#dtanh/dz  = 1-a^2
        else:
            dZL = np.dot(parameters["W"+str(L+1)].T,dZL)* np.array(cache["Z"+str(L)]>0) 
        
        gradient["dW"+str(L)] = (1/m)*np.dot(dZL,cache["A"+str(L-1)].T)
        gradient["db"+str(L)] = (1/m)*np.sum(dZL,axis=1,keepdims = True)        
    return gradient

def update_parameters_with_momentum(gradient,parameters,LearnRate,v,momentum=0.9):
    Layer_num = len(parameters) // param_num
    for L in range(1,Layer_num+1): #遍历[1,2,3,4]
        #["dW" + str(L)] = momentum * v["dW" + str(L)] + (1 - momentum) * gradient["dW" + str(L)]
        v["dW" + str(L)] = momentum * v["dW" + str(L)] + gradient["dW" + str(L)]
        parameters["W"+str(L)] = parameters["W"+str(L)] - LearnRate*v["dW"+str(L)]
        #["db" + str(L)] = momentum * v["db" + str(L)] + (1 - momentum) * gradient["db" + str(L)]
        v["db" + str(L)] = momentum * v["db" + str(L)] + gradient["db" + str(L)]
        parameters["b"+str(L)] = parameters["b"+str(L)] - LearnRate*v["db"+str(L)]
    return parameters,v

def update_parameters(gradient,parameters,LearnRate):
    #w:=w-lr*dw  ;   b:=b-lr*db
    Layer_num = len(parameters) // param_num
    for L in range(1,Layer_num+1): #遍历[1,2,3,4]
        parameters["W"+str(L)] = parameters["W"+str(L)] - LearnRate*gradient["dW"+str(L)]
        parameters["b"+str(L)] = parameters["b"+str(L)] - LearnRate*gradient["db"+str(L)]   
    return parameters
              
def Train_Net(fc_net,train_set_x,train_set_y,check=False,iterations=2000,LearnRate=0.01,active_func="ReLU"):
    #4.初始化参数
    #arameters = init_parameters(fc_net)  
    parameters,v = init_parameters_momentum(fc_net)
    costs = [] 
    for iteration in range(0,iterations):
        AL,cache= forward_pass(train_set_x,parameters,active_func) #AL=(1,209)
        #6.计算代价值
        cost = compute_cost(AL,train_set_y)         
        if iteration%500 == 0: 
            print("iteration = ",iteration,";    cost = ",cost)
            costs.append(cost)
        #7.反向传播计算梯度
        gradient = backward_pass(AL,parameters,cache,train_set_y,active_func)
        #8.根据梯度更新一次参数
        #arameters = update_parameters(gradient,parameters,LearnRate)
        parameters,v = update_parameters_with_momentum(gradient,parameters,LearnRate,v)
           
    plt.plot(costs,'r')
    plt.xlabel("iteration")
    plt.ylabel("cost")
    plt.show()
    return parameters

def Predict(A0,Y,parameters,active_func):
    AL,_ = forward_pass(A0,parameters,active_func)#AL是（1,50）
    m = AL.shape[1]
    p = np.zeros(AL.shape)
    for i in range(0,AL.shape[1]):
        if AL[0,i]>0.5:
            p[0,i]=1
        else:
            p[0,i]=0           
    accuracy = (1/m)* np.sum(p==Y)
    print("accuracy =",accuracy)
            
    
if __name__ == '__main__':
    #1.加载数据
    train_set_x,train_set_y,test_set_x,test_set_y = load_dataset()
    print("train_set_x.shape=",train_set_x.shape,"    train_set_y.shape=",train_set_y.shape)
    print("test_set_x.shape=",test_set_x.shape,"    test_set_y.shape=",test_set_y.shape)

    #2.对输入像素值做归一化（0~255）->(0~1)
    train_set_x = train_set_x/255.
    test_set_x = test_set_x/255.
    #3.定义全连接神经网络各层神经元个数，并初始化参数w和b
    fc_net = [12288,10,3,2,1]       #四层网络（梯度消失）
    #4.开始训练
    parameters = Train_Net(fc_net,train_set_x,train_set_y,iterations=3001,LearnRate=0.1,active_func="ReLU") 
    #5.开始预测
    print("===Predict Train Set ===")
    Predict(train_set_x,train_set_y,parameters,active_func="ReLU")
    print("===Predict Test Set ===")
    Predict(test_set_x,test_set_y,parameters,active_func="ReLU")
    
    
    
#把数据重新打乱顺序，进行输入，可以让数据更加具备典型性和更优良的泛化能力！
#对输入顺序进行随机化处理是为了保证能够有监督学习，同时使算法按照梯度下降法则进行学习。
#假如训练数据是有序的，那么会导致训练结果很难收敛到偏置值。只有保证数据的随机性才能使得BP算法训练结果尽可能地收敛。  
#打乱数据集:要确保特征和标签都被同步地重新排列。
"""
这来自于信息理论（Information Theory）——“学习到一件不太可能发生的事却发生了，比学习一件很可能发生的事已经发生，包含更多的信息。”同样的，
把训练样例的顺序随机化（在不同周期，或者 mini-batch），会导致更快的收敛。如果模型看到的很多样例不在同一种顺序下，运算速度会有小幅提升。

"""
def random_mini_batches(X, Y, mini_batch_size = 64):
    m = X.shape[1]                  # number of training examples
    mini_batches = []        
    # Step 1: 打乱数据集
    permutation = list(np.random.permutation(m))#获取0~m-1的随机排列数,这些随机排列数作为列索引，就能把batch打乱了（必要的步骤）
    shuffled_X = X[:, permutation] 
    shuffled_Y = Y[:, permutation].reshape((1,m))
    # Step 2: 分组 (shuffled_X, shuffled_Y)
    num_complete_minibatches = math.floor(m/mini_batch_size) # number of mini batches of size mini_batch_size in your partitionning
    for k in range(0, num_complete_minibatches):
        mini_batch_X = shuffled_X[:, k * mini_batch_size : (k+1) * mini_batch_size]
        mini_batch_Y = shuffled_Y[:, k * mini_batch_size : (k+1) * mini_batch_size]
        mini_batch = (mini_batch_X, mini_batch_Y)
        mini_batches.append(mini_batch)  
    if m % mini_batch_size != 0:
        mini_batch_X = shuffled_X[:, (k+1) * mini_batch_size :]
        mini_batch_Y = shuffled_Y[:, (k+1) * mini_batch_size :]
        mini_batch = (mini_batch_X, mini_batch_Y)
        mini_batches.append(mini_batch)    
    return mini_batches
    
#8001   cost =  0.000935400328228   accuracy = 0.76 (gd+relu)
#8001   cost =  0.000540447819058   accuracy = 0.72 (momentum+relu)->数据量太少，过拟合

#iteration =  7000 ;    cost =  0.000439735530467  accuracy = 0.70 (gd+relu)
#7001   cost =  0.0290159490925   accuracy = 0.62 (momentum+relu)   
#多次试验，在本例中，gd和momentum各有胜负，可能是数据量太小，看不出太明显的区别来
import time
start =time.clock()

#中间写上代码块

end = time.clock()
print('Running time: %s Seconds'%(end-start))














def update_parameters_with_momentum(gradient,parameters,LearnRate,v,momentum=0.9):
    Layer_num = len(parameters) // 2
    for L in range(1,Layer_num+1): #遍历[1,2,3,4]
        #["dW" + str(L)] = momentum * v["dW" + str(L)] + (1 - momentum) * gradient["dW" + str(L)]
        v["dW" + str(L)] = momentum * v["dW" + str(L)] + gradient["dW" + str(L)]
        parameters["W"+str(L)] = parameters["W"+str(L)] - LearnRate*v["dW"+str(L)]
        #["db" + str(L)] = momentum * v["db" + str(L)] + (1 - momentum) * gradient["db" + str(L)]
        v["db" + str(L)] = momentum * v["db" + str(L)] + gradient["db" + str(L)]
        parameters["b"+str(L)] = parameters["b"+str(L)] - LearnRate*v["db"+str(L)]
    return parameters,v
	
def init_parameters_momentum(fc_net):
    parameters = {}
    v = {}
    Layer_num = len(fc_net)     
    for L in range(1,Layer_num):
        #parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*0.01   #最普通的随机初始化
        parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])* np.sqrt(2/fc_net[L-1])
        parameters["b"+str(L)] = np.zeros((fc_net[L],1))  #使用BN的时候其实可以去掉偏置b了（要在很多代码里面改，先不改了......）     
        v["dW" + str(L)] = np.zeros((fc_net[L],fc_net[L-1]))
        v["db" + str(L)] = np.zeros((fc_net[L],1))
    for L in range(1,Layer_num):
        print("W"+str(L)+"=",parameters["W"+str(L)].shape)
        print("b"+str(L)+"=",parameters["b"+str(L)].shape)                       
    return parameters,v
	
	
	
	
	
	
	
def init_parameters_RMSProp(fc_net):
    parameters = {}
    Layer_num = len(fc_net) #Layer_num=5 
    s = {}
    for L in range(1,Layer_num):
        #parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*0.01  #
        #parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*np.sqrt(1/fc_net[L-1]) #Xavier初始化，针对tanh函数
        parameters["W"+str(L)] = np.random.randn(fc_net[L],fc_net[L-1])*np.sqrt(2/fc_net[L-1]) #He初始化，针对ReLU函数
        parameters["b"+str(L)] = np.zeros((fc_net[L],1))     
        s["dW"+str(L)] = np.zeros((fc_net[L],fc_net[L-1]))
        s["db"+str(L)] = np.zeros((fc_net[L],1))           
    for L in range(1,Layer_num):
        print("W"+str(L)+"=",parameters["W"+str(L)].shape)
        print("b"+str(L)+"=",parameters["b"+str(L)].shape)                       
    return parameters,s

	
def update_parameters_with_RMSProp(gradient,parameters,LearnRate,s,decay=0.9,eps=1e-8):
    #w:=w-lr*dw  ;   b:=b-lr*db
    Layer_num = len(parameters) // 2
    for L in range(1,Layer_num+1): #遍历[1,2,3,4]
        s["dW"+str(L)] = decay*s["dW"+str(L)] + (1-decay)*(gradient["dW"+str(L)]**2)    
        parameters["W"+str(L)] = parameters["W"+str(L)] - LearnRate*gradient["dW"+str(L)]/np.sqrt(s["dW"+str(L)]+eps)
        s["db"+str(L)] = decay*s["db"+str(L)] + (1-decay)*(gradient["db"+str(L)]**2)
        parameters["b"+str(L)] = parameters["b"+str(L)] - LearnRate*gradient["db"+str(L)]/np.sqrt(s["db"+str(L)]+eps)
    return parameters,s
